# import numpy as np
# import pytest
# from test_inputs.catchment import *
# from test_inputs.readers import *

# import earthkit.hydro as ekh


# @pytest.mark.parametrize(
#     "river_network",
#     [
#         ("d8_ldd", d8_ldd_1),
#         ("cama_downxy", cama_downxy_1),
#         ("cama_nextxy", cama_nextxy_1),
#         ("d8_ldd", d8_ldd_2),
#         ("cama_downxy", cama_downxy_2),
#         ("cama_nextxy", cama_nextxy_2),
#     ],
#     indirect=True,
# )
# def test_find_subcatchments_does_not_overwrite(river_network):
#     field = np.arange(river_network.n_nodes) + 1
#     subcatchments = ekh.subcatchments.find(river_network, field)
#     print(subcatchments)
#     print(field)
#     np.testing.assert_array_equal(subcatchments, field)


# @pytest.mark.parametrize(
#     "river_network",
#     [
#         ("d8_ldd", d8_ldd_1),
#         ("cama_downxy", cama_downxy_1),
#         ("cama_nextxy", cama_nextxy_1),
#         ("d8_ldd", d8_ldd_2),
#         ("cama_downxy", cama_downxy_2),
#         ("cama_nextxy", cama_nextxy_2),
#     ],
#     indirect=True,
# )
# def test_find_subcatchments_does_not_overwrite_2d(river_network):
#     field = np.zeros(river_network.mask.shape, dtype="int")
#     field[river_network.mask] = np.arange(river_network.n_nodes) + 1
#     find_subcatchments = ekh.subcatchments.find(river_network, field)
#     print(find_subcatchments)
#     print(field)
#     np.testing.assert_array_equal(find_subcatchments, field)


# @pytest.mark.parametrize(
#     "river_network, query_field, subcatchment",
#     [
#         (("d8_ldd", d8_ldd_1), catchment_query_field_1, subcatchment_1),
#         (("cama_downxy", cama_downxy_1), catchment_query_field_1, subcatchment_1),
#         (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, subcatchment_1),
#         (("d8_ldd", d8_ldd_2), catchment_query_field_2, subcatchment_2),
#         (("cama_downxy", cama_downxy_2), catchment_query_field_2, subcatchment_2),
#         (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, subcatchment_2),
#     ],
#     indirect=["river_network"],
# )
# def test_find_subcatchments(river_network, query_field, subcatchment):
#     subcatchments = ekh.subcatchments.find(river_network, query_field)
#     print(subcatchment)
#     print(subcatchments)
#     np.testing.assert_array_equal(subcatchment, subcatchments)


# @pytest.mark.parametrize(
#     "river_network, query_field, find_subcatchments",
#     [
#         (("d8_ldd", d8_ldd_1), catchment_query_field_1, subcatchment_1),
#         (("cama_downxy", cama_downxy_1), catchment_query_field_1, subcatchment_1),
#         (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, subcatchment_1),
#         (("d8_ldd", d8_ldd_2), catchment_query_field_2, subcatchment_2),
#         (("cama_downxy", cama_downxy_2), catchment_query_field_2, subcatchment_2),
#         (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, subcatchment_2),
#     ],
#     indirect=["river_network"],
# )
# def test_find_subcatchments_2d(river_network, query_field, find_subcatchments):
#     field = np.zeros(river_network.mask.shape, dtype="int")
#     field[river_network.mask] = query_field
#     network_find_subcatchments = ekh.subcatchments.find(river_network, field)
#     print(find_subcatchments)
#     print(network_find_subcatchments)
#     np.testing.assert_array_equal(
#         network_find_subcatchments[river_network.mask], find_subcatchments
#     )
#     np.testing.assert_array_equal(network_find_subcatchments[~river_network.mask], 0)


# @pytest.mark.parametrize(
#     "river_network, query_field, find_catchments",
#     [
#         (("d8_ldd", d8_ldd_1), catchment_query_field_1, catchment_1),
#         (("cama_downxy", cama_downxy_1), catchment_query_field_1, catchment_1),
#         (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, catchment_1),
#         (("d8_ldd", d8_ldd_2), catchment_query_field_2, catchment_2),
#         (("cama_downxy", cama_downxy_2), catchment_query_field_2, catchment_2),
#         (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, catchment_2),
#     ],
#     indirect=["river_network"],
# )
# def test_find_catchments_2d(river_network, query_field, find_catchments):
#     field = np.zeros(river_network.mask.shape, dtype="int")
#     field[river_network.mask] = query_field
#     network_find_catchments = ekh.catchments.find(river_network, field)
#     print(find_catchments)
#     print(network_find_catchments)
#     np.testing.assert_array_equal(
#         network_find_catchments[river_network.mask], find_catchments
#     )
#     np.testing.assert_array_equal(network_find_catchments[~river_network.mask], 0)



def _ufunc_to_downstream(
    field,
    did,
    uid,
    eid,
    node_additive_weight,
    node_multiplicative_weight,
    node_modifier_use_upstream,
    edge_additive_weight,
    edge_multiplicative_weight,
    func,
    xp,
):
    """
    Updates field in-place by applying a ufunc at the downstream nodes
    of the grouping.

    Parameters
    ----------
    river_network : earthkit.hydro.network.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    grouping : numpy.ndarray
        An array of indices.
    mv : scalar
        A missing value indicator (not used in the function but kept for consistency).
    additive_weight : numpy.ndarray, optional
        A weight to be added to the field values. Default is None.
    multiplicative_weight : numpy.ndarray, optional
        A weight to be multiplied with the field values. Default is None.
    modifier_use_upstream : bool, optional
        If True, the modifiers are used on the upstream nodes instead of downstream.
        Default is True.
    ufunc : numpy.ufunc
        A universal function from the numpy library to be applied to the field data.
        Available ufuncs: https://numpy.org/doc/2.2/reference/ufuncs.html.
        Note: must allow two operands.

    Returns
    -------
    None
    """
    modifier_group = uid if node_modifier_use_upstream else did

    modifier_field = xp.gather(field, uid, axis=-1)
    if node_multiplicative_weight is not None:
        modifier_field *= node_multiplicative_weight[..., modifier_group]
    if edge_multiplicative_weight is not None:
        modifier_field *= edge_multiplicative_weight[..., eid]
    if node_additive_weight is not None:
        modifier_field += node_additive_weight[..., modifier_group]
    if edge_additive_weight is not None:
        modifier_field += edge_additive_weight[..., eid]

    return func(
        field,
        did,
        modifier_field,
    )



import numpy as np

from earthkit.hydro.data_structures import RiverNetwork


def propagate(
    river_network: RiverNetwork,
    groups: np.ndarray,
    field: np.ndarray,
    invert_graph: bool,
    operation,
    *args,
    **kwargs,
):
    if invert_graph:
        for uid, did, eid in groups[::-1]:
            field = operation(field, did, uid, eid, *args, **kwargs)
    else:
        for did, uid, eid in groups:
            field = operation(field, did, uid, eid, *args, **kwargs)

    return field



import numpy as np
from earthkit.utils.array import to_device

from ._network_storage import RiverNetworkStorage


class RiverNetwork:
    """
    A class representing a river network for hydrological processing.

    Attributes
    ----------
    n_nodes : int
        The number of nodes in the river network.
    n_edges : int
        The number of nodes in the river network.
    sinks : array-like
        Nodes with no downstream connections.
    sources : array-like
        Nodes with no upstream connections.
    bifurcates : bool
        Whether the river network has bifurcations.
    shape : tuple
        The size of the river network grid. None if the network is vector-based.
    mask : array-like
        Flattened 1D indices on the raster grid corresponding to river network nodes.
    array_backend : str
        The array backend of the river network.
    device : str
        The device of the river network.
    return_type : str
        The default return type of the river network. Either "gridded" or "masked".
    """

    def __init__(self, river_network_storage: RiverNetworkStorage):
        self._storage = river_network_storage
        self.n_nodes = self._storage.n_nodes
        self.n_edges = self._storage.n_edges
        self.sources = self._storage.sources
        self.sinks = self._storage.sinks

        self.bifurcates = self._storage.bifurcates
        self.edge_weights = self._storage.edge_weights

        self.mask = self._storage.mask
        self.shape = self._storage.shape
        self.array_backend = "numpy"
        self.device = "cpu"
        self.return_type = "gridded"

        self.coords = self._storage.coords

        self.data = [self._storage.sorted_data]
        self.groups = np.split(self._storage.sorted_data, self._storage.splits, axis=1)

    def __str__(self):
        return f"RiverNetwork with {self.n_nodes} nodes and {self.n_edges} edges."

    def __repr__(self):
        return self.__str__()

    def to_device(self, device=None, array_backend=None):
        """
        Change the RiverNetwork's array backend and/or move it to a
        different device.

        Parameters
        ----------
        device : str, optional
            The device to which to transfer. Default is None, which is `'cpu'` for all backends except cupy, which is `'gpu'`.
        array_backend : str, optional
            The array backend.
            One of "numpy", "np", "cupy", "cp", "pytorch", "torch", "jax", "jnp", "tensorflow" or "tf".
            Default is None, which uses `self.array_backend`.

        Returns
        -------
        RiverNetwork
            The modified RiverNetwork.
        """

        # TODO: use xp.asarray
        if array_backend == "np":
            array_backend = "numpy"
        elif array_backend == "cp":
            array_backend = "cupy"
        elif array_backend == "jnp":
            array_backend = "jax"
        elif array_backend == "tf":
            array_backend = "tensorflow"
        elif array_backend == "pytorch":
            array_backend = "torch"

        if device is None:
            device = "cpu" if array_backend != "cupy" else "gpu"
        if array_backend is None:
            if self.array_backend == "numpy" and device in ["gpu", "cuda"]:
                array_backend = "cupy"
            else:
                array_backend = self.array_backend

        if array_backend in ["torch", "cupy", "numpy"]:
            self.groups = [
                to_device(group, device, array_backend=array_backend)
                for group in self.groups
            ]
            self.mask = to_device(self.mask, device, array_backend=array_backend)
            self.data = [to_device(self.data[0], device, array_backend=array_backend)]
        elif array_backend == "jax":
            assert device == "cpu"
            import jax.numpy as jnp

            self.groups = [jnp.array(x) for x in self.groups]
            self.mask = jnp.array(self.mask)
            self.data = [jnp.array(self.data[0])]
        elif array_backend == "tensorflow":
            assert device == "cpu"
            import tensorflow as tf

            self.groups = [tf.convert_to_tensor(x, dtype=tf.int32) for x in self.groups]
            self.mask = tf.convert_to_tensor(self.mask, dtype=tf.int32)
            self.data = [tf.convert_to_tensor(self.data[0], dtype=tf.int32)]
        else:
            raise NotImplementedError

        self.array_backend = array_backend
        self.device = device
        return self

    def set_default_return_type(self, return_type):
        """
        Set the default return type for the river network.

        Parameters
        ----------
        return_type : str
            The default return_type to use.

        Returns
        -------
        None
        """
        if return_type not in ["gridded", "masked"]:
            raise ValueError(
                f'Invalid return_type {return_type}. Valid types are "gridded", "masked"'
            )
        self.return_type = return_type

    def export(self, fpath="river_network.joblib", compression=1):
        """
        Save the river network to a local file.

        Parameters
        ----------
        fpath : str, optional
            The filepath specifying where to save the RiverNetwork. Default is `'river_network.joblib'`.
        compression : str, optional
            The compression factor used for saving. Default is 1.

        Returns
        -------
        None
        """
        import joblib

        joblib.dump(self._storage, fpath, compress=compression)



def get_array_backend(x):

    if type(x) is str:
        mod = x
    else:
        mod = type(x).__module__

    if "torch" in mod:
        from .torch_backend import TorchBackend

        return TorchBackend()
    elif "tensorflow" in mod:
        from .tensorflow_backend import TFBackend

        return TFBackend()
    elif "jax" in mod:
        from .jax_backend import JAXBackend

        return JAXBackend()
    elif "cupy" in mod:
        from .cupy_backend import CuPyBackend

        return CuPyBackend()
    elif "numpy" in mod:
        from .numpy_backend import NumPyBackend

        return NumPyBackend()
    else:
        raise TypeError(f"Unsupported array type: {type(x)}")



from earthkit.hydro.catchments.array import _operations


def var(river_network, field, locations, node_weights=None, edge_weights=None):
    r"""
    Computes the weighted variance of a field over the upstream
    catchment of each specified location.

    For each location, this function identifies all upstream nodes in the river network
    and accumulates their contributions downstream, weighted by both node and edge weights.

    The weighted variance is defined as:

    .. math::
       :nowrap:

       \begin{align*}
       x'_i &= w'_i \cdot x_i \\
       q'_i &= w'_i \cdot x_i^2 \\
       n_j &= x'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot n_i \\
       q_j &= q'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot q_i \\
       d_j &= w'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot d_i \\
       \bar{x}_j &= \frac{n_j}{d_j} \\
       \mathrm{Var}(x)_j &= \frac{q_j}{d_j} - \bar{x}_j^2
       \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g., discharge partitioning ratio),
    - :math:`\mathrm{Up}(j)` is the set of upstream nodes flowing into node :math:`j`,
    - :math:`n_j` is the accumulated weighted value,
    - :math:`q_j` is the accumulated weighted squared value,
    - :math:`d_j` is the accumulated weight (denominator),
    - :math:`\bar{x}_j` is the weighted average at node :math:`j`,
    - :math:`\mathrm{Var}(x)_j` is the weighted variance at node :math:`j`.

    Accumulation proceeds in topological order from the sources to the sinks. This formulation computes the population variance.


    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    locations : array-like or dict
        A list of nodes at which to compute.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each river network edge. Default is None (unweighted).

    Returns
    -------
    array-like
        Array of variance values for each location in `locations`.
    """
    return _operations.var(river_network, field, locations, node_weights, edge_weights)


def std(river_network, field, locations, node_weights=None, edge_weights=None):
    r"""
    Computes the weighted standard deviation of a field over the
    upstream catchment of each specified location.

    For each location, this function identifies all upstream nodes in the river network
    and accumulates their contributions downstream, weighted by both node and edge weights.

    The weighted standard deviation is defined as:

    .. math::
       :nowrap:

       \begin{align*}
       x'_i &= w'_i \cdot x_i \\
       q'_i &= w'_i \cdot x_i^2 \\
       n_j &= x'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot n_i \\
       q_j &= q'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot q_i \\
       d_j &= w'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot d_i \\
       \bar{x}_j &= \frac{n_j}{d_j} \\
       \mathrm{Var}(x)_j &= \frac{q_j}{d_j} - \bar{x}_j^2 \\
       \mathrm{Std}(x)_j &= \sqrt{\mathrm{Var}(x)_j}
       \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g., discharge partitioning ratio),
    - :math:`\mathrm{Up}(j)` is the set of upstream nodes flowing into node :math:`j`,
    - :math:`n_j` is the accumulated weighted value,
    - :math:`q_j` is the accumulated weighted squared value,
    - :math:`d_j` is the accumulated weight (denominator),
    - :math:`\bar{x}_j` is the weighted average at node :math:`j`,
    - :math:`\mathrm{Var}(x)_j` is the weighted variance at node :math:`j`.
    - :math:`\mathrm{Std}(x)_j` is the weighted standard deviation at node :math:`j`.

    Accumulation proceeds in topological order from the sources to the sinks. This formulation computes the population standard deviation.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    locations : array-like or dict
        A list of nodes at which to compute.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each river network edge. Default is None (unweighted).

    Returns
    -------
    array-like
        Array of standard deviation values for each location in `locations`.
    """
    return _operations.std(river_network, field, locations, node_weights, edge_weights)


def mean(river_network, field, locations, node_weights=None, edge_weights=None):
    r"""
    Computes the weighted mean of a field over the upstream catchment of
    each specified location.

    For each location, this function identifies all upstream nodes in the river network
    and accumulates their contributions downstream, weighted by both node and edge weights.

    The weighted mean is defined as:

    .. math::
       :nowrap:

       \begin{align*}
       x'_i &= w'_i \cdot x_i \\
       n_j &= x'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot n_i \\
       d_j &= w'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot d_i \\
       \bar{x}_j &= \frac{n_j}{d_j}
       \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Up}(j)` is the set of upstream nodes flowing into node :math:`j`,
    - :math:`n_j` is the accumulated weighted value,
    - :math:`d_j` is the accumulated weight (denominator),
    - :math:`\bar{x}_j` is the weighted mean at node :math:`j`.

    Accumulation proceeds in topological order from the sources to the sinks.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    locations : array-like or dict
        A list of nodes at which to compute.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each river network edge. Default is None (unweighted).

    Returns
    -------
    array-like
        Array of mean values for each location in `locations`.
    """
    return _operations.mean(river_network, field, locations, node_weights, edge_weights)


def sum(river_network, field, locations, node_weights=None, edge_weights=None):
    r"""
    Computes the weighted sum of a field over the upstream catchment of
    each specified location.

    For each location, this function identifies all upstream nodes in the river network
    and accumulates their contributions downstream, weighted by both node and edge weights.

    The weighted sum is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        n_j &= x'_j + \sum_{i \in \mathrm{Up}(j)} w_{ij} \cdot n_i
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Up}(j)` is the set of upstream nodes flowing into node :math:`j`,
    - :math:`n_j` is the weighted sum at node :math:`j`.

    Accumulation proceeds in topological order from the sources to the sinks.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    locations : array-like or dict
        A list of nodes at which to compute.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each river network edge. Default is None (unweighted).

    Returns
    -------
    array-like
        Array of sum values for each location in `locations`.
    """
    return _operations.sum(river_network, field, locations, node_weights, edge_weights)


def min(river_network, field, locations, node_weights=None, edge_weights=None):
    r"""
    Computes the weighted minimum of a field over the upstream catchment
    of each specified location.

    For each location, this function identifies all upstream nodes in the river network
    and accumulates their contributions downstream, weighted by both node and edge weights.

    The weighted minimum is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        m_j &= \mathrm{min}(x'_j,~\mathrm{min}_{i \in \mathrm{Up}(j)} w_{ij} \cdot m_i)
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Up}(j)` is the set of upstream nodes flowing into node :math:`j`,
    - :math:`m_j` is the weighted minimum at node :math:`j`.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    locations : array-like or dict
        A list of nodes at which to compute.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each river network edge. Default is None (unweighted).

    Returns
    -------
    array-like
        Array of minimum values for each location in `locations`.
    """
    return _operations.min(river_network, field, locations, node_weights, edge_weights)


def max(river_network, field, locations, node_weights=None, edge_weights=None):
    r"""
    Computes the weighted maximum of a field over the upstream catchment
    of each specified location.

    For each location, this function identifies all upstream nodes in the river network
    and accumulates their contributions downstream, weighted by both node and edge weights.

    The weighted maximum is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        m_j &= \mathrm{max} (x'_j,~\mathrm{max}_{i \in \mathrm{Up}(j)} w_{ij} \cdot m_i)
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Up}(j)` is the set of upstream nodes flowing into node :math:`j`,
    - :math:`m_j` is the weighted maximum at node :math:`j`.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    locations : array-like or dict
        A list of nodes at which to compute.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each river network edge. Default is None (unweighted).

    Returns
    -------
    array-like
        Array of maximum values for each location in `locations`.
    """
    return _operations.max(river_network, field, locations, node_weights, edge_weights)


def find(river_network, locations, overwrite=True, return_type=None):
    r"""
    Delineates catchment areas.

    Given a field indicating one or more start locations (e.g., outlet points or pour points),
    this function delineates the catchments upstream of each start location by grouping all cells that flow into these points.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    locations : array-like or dict
        A list of catchment sink nodes (start locations).
    overwrite : bool, optional
        Whether to overwrite subcatchments or not. Default is True.
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.

    Returns
    -------
    array-like
        Array of labelled catchments for every river network node or gridcell, depending on `return_grid`.
    """
    return _operations.find(river_network, locations, overwrite, return_type)



from earthkit.hydro.downstream.array import _operations


def var(river_network, field, node_weights=None, edge_weights=None, return_type=None):
    r"""
    Computes the weighted variance of a field over all downstream nodes.

    For each node in the river network, this function identifies all downstream nodes in the river network
    and accumulates their contributions upstream, weighted by both node and edge weights.

    The weighted variance is defined as:

    .. math::
       :nowrap:

       \begin{align*}
       x'_i &= w'_i \cdot x_i \\
       q'_i &= w'_i \cdot x_i^2 \\
       n_j &= x'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot n_i \\
       q_j &= q'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot q_i \\
       d_j &= w'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot d_i \\
       \bar{x}_j &= \frac{n_j}{d_j} \\
       \mathrm{Var}(x)_j &= \frac{q_j}{d_j} - \bar{x}_j^2
       \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g., discharge partitioning ratio),
    - :math:`\mathrm{Down}(j)` is the set of downstream nodes flowing out of node :math:`j`,
    - :math:`n_j` is the accumulated weighted value,
    - :math:`q_j` is the accumulated weighted squared value,
    - :math:`d_j` is the accumulated weight (denominator),
    - :math:`\bar{x}_j` is the weighted average at node :math:`j`,
    - :math:`\mathrm{Var}(x)_j` is the weighted variance at node :math:`j`.

    Accumulation proceeds in inverse topological order from the sinks to the sources. This formulation computes the population variance.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.

    Returns
    -------
    array-like
        Array of variance values for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.var(
        river_network, field, node_weights, edge_weights, return_type
    )


def std(river_network, field, node_weights=None, edge_weights=None, return_type=None):
    r"""
    Computes the weighted standard deviation of a field over all
    downstream nodes.

    For each node in the river network, this function identifies all downstream nodes in the river network
    and accumulates their contributions upstream, weighted by both node and edge weights.

    The weighted standard deviation is defined as:

    .. math::
       :nowrap:

       \begin{align*}
       x'_i &= w'_i \cdot x_i \\
       q'_i &= w'_i \cdot x_i^2 \\
       n_j &= x'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot n_i \\
       q_j &= q'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot q_i \\
       d_j &= w'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot d_i \\
       \bar{x}_j &= \frac{n_j}{d_j} \\
       \mathrm{Var}(x)_j &= \frac{q_j}{d_j} - \bar{x}_j^2 \\
       \mathrm{Std}(x)_j &= \sqrt{\mathrm{Var}(x)_j}
       \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g., discharge partitioning ratio),
    - :math:`\mathrm{Down}(j)` is the set of downstream nodes flowing out of node :math:`j`,
    - :math:`n_j` is the accumulated weighted value,
    - :math:`q_j` is the accumulated weighted squared value,
    - :math:`d_j` is the accumulated weight (denominator),
    - :math:`\bar{x}_j` is the weighted average at node :math:`j`,
    - :math:`\mathrm{Var}(x)_j` is the weighted variance at node :math:`j`.
    - :math:`\mathrm{Std}(x)_j` is the weighted standard deviation at node :math:`j`.

    Accumulation proceeds in inverse topological order from the sinks to the sources. This formulation computes the population standard deviation.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.

    Returns
    -------
    array-like
        Array of standard deviation values for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.std(
        river_network, field, node_weights, edge_weights, return_type
    )


def mean(river_network, field, node_weights=None, edge_weights=None, return_type=None):
    r"""
    Computes the weighted mean of a field over all downstream nodes.

    For each node in the river network, this function identifies all downstream nodes in the river network
    and accumulates their contributions upstream, weighted by both node and edge weights.

    The weighted mean is defined as:

    .. math::
       :nowrap:

       \begin{align*}
       x'_i &= w'_i \cdot x_i \\
       n_j &= x'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot n_i \\
       d_j &= w'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot d_i \\
       \bar{x}_j &= \frac{n_j}{d_j}
       \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Down}(j)` is the set of downstream nodes flowing out of node :math:`j`,
    - :math:`n_j` is the accumulated weighted value,
    - :math:`d_j` is the accumulated weight (denominator),
    - :math:`\bar{x}_j` is the weighted mean at node :math:`j`.

    Accumulation proceeds in inverse topological order from the sinks to the sources.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.

    Returns
    -------
    array-like
        Array of mean values for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.mean(
        river_network, field, node_weights, edge_weights, return_type
    )


def sum(river_network, field, node_weights=None, edge_weights=None, return_type=None):
    r"""
    Computes the weighted sum of a field over all downstream nodes.

    For each node in the river network, this function identifies all downstream nodes in the river network
    and accumulates their contributions upstream, weighted by both node and edge weights.

    The weighted sum is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        n_j &= x'_j + \sum_{i \in \mathrm{Down}(j)} w_{ij} \cdot n_i
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Down}(j)` is the set of downstream nodes flowing out of node :math:`j`,
    - :math:`n_j` is the weighted sum at node :math:`j`.

    Accumulation proceeds in inverse topological order from the sinks to the sources.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.

    Returns
    -------
    array-like
        Array of sum values for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.sum(
        river_network, field, node_weights, edge_weights, return_type
    )


def min(river_network, field, node_weights=None, edge_weights=None, return_type=None):
    r"""
    Computes the weighted minimum of a field over all downstream nodes.

    For each node in the river network, this function identifies all downstream nodes in the river network
    and accumulates their contributions upstream, weighted by both node and edge weights.

    The weighted minimum is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        m_j &= \mathrm{min}(x'_j,~\mathrm{min}_{i \in \mathrm{Down}(j)} w_{ij} \cdot m_i)
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Down}(j)` is the set of downstream nodes flowing out of node :math:`j`,
    - :math:`m_j` is the weighted minimum at node :math:`j`.

    Accumulation proceeds in inverse topological order from the sinks to the sources.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.

    Returns
    -------
    array-like
        Array of minimum values for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.min(
        river_network, field, node_weights, edge_weights, return_type
    )


def max(river_network, field, node_weights=None, edge_weights=None, return_type=None):
    r"""
    Computes the weighted maximum of a field over all downstream nodes.

    For each node in the river network, this function identifies all downstream nodes in the river network
    and accumulates their contributions upstream, weighted by both node and edge weights.

    The weighted maximum is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        m_j &= \mathrm{max} (x'_j,~\mathrm{max}_{i \in \mathrm{Down}(j)} w_{ij} \cdot m_i)
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Down}(j)` is the set of downstream nodes flowing out of node :math:`j`,
    - :math:`m_j` is the weighted maximum at node :math:`j`.

    Accumulation proceeds in inverse topological order from the sinks to the sources.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.

    Returns
    -------
    array-like
        Array of maximum values for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.max(
        river_network, field, node_weights, edge_weights, return_type
    )



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import earthkit.hydro.catchments
import earthkit.hydro.distance
import earthkit.hydro.downstream
import earthkit.hydro.length
import earthkit.hydro.move
import earthkit.hydro.river_network
import earthkit.hydro.subnetwork
import earthkit.hydro.upstream

from ._version import __version__

__all__ = [
    "catchments",
    "distance",
    "downstream",
    "length",
    "move",
    "river_network",
    "upstream",
    "subnetwork",
    "__version__",
]



from functools import wraps
from inspect import signature

import numpy as np
import xarray as xr

from earthkit.hydro._utils.coords import get_core_dims, node_default_coord


def get_full_signature(func, *args, **kwargs):
    sig = signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    return bound_args.arguments


def assert_xr_compatible_backend(network):
    network_backend = network.array_backend
    if network_backend not in ["numpy", "cupy"]:
        raise NotImplementedError(f"xarray does not support {network_backend} backend")


def sort_xr_nonxr_args(all_args):
    xr_args = []
    non_xr_kwargs = {}
    arg_order = []
    for name, value in all_args.items():
        if isinstance(value, (xr.DataArray, xr.Dataset)):
            xr_args.append(value)
            arg_order.append(("xr", name))
        else:
            non_xr_kwargs[name] = value
            arg_order.append(("nonxr", name))
    return xr_args, non_xr_kwargs, arg_order


def get_reshuffled_func(func, arg_order):
    def reshuffled_func(*only_xr_args, **non_xr_kwargs):
        full_args = {}
        xr_i = 0
        for kind, name in arg_order:
            if kind == "xr":
                full_args[name] = only_xr_args[xr_i]
                xr_i += 1
            else:
                full_args[name] = non_xr_kwargs[name]
        return func(**full_args)

    return reshuffled_func


def get_input_output_core_dims(
    input_core_dims, output_core_dims, xr_args, river_network, return_grid
):
    if input_core_dims is None:
        input_core_dims = [get_core_dims(xr_arg) for xr_arg in xr_args]
    elif len(input_core_dims) == 1:
        input_core_dims *= len(xr_args)

    if output_core_dims is None:
        if return_grid:
            if len(input_core_dims[0]) == 2:  # grid in and out
                output_core_dims = [input_core_dims[0]]
            else:
                output_core_dims = [
                    list(river_network.coords.keys())
                ]  # 1d in, grid out
        else:
            if len(input_core_dims[0]) == 1:  # 1d in and out
                output_core_dims = [input_core_dims[0]]
            else:
                output_core_dims = [[node_default_coord]]

    return input_core_dims, output_core_dims


def xarray(func):

    @wraps(func)
    def wrapper(*args, **kwargs):

        # Inspect the function signature and bind all arguments
        all_args = get_full_signature(func, *args, **kwargs)

        input_core_dims = all_args.pop("input_core_dims", None)
        output_core_dims = None

        assert_xr_compatible_backend(all_args["river_network"])

        # Separate xarray and non-xarray arguments
        xr_args, non_xr_kwargs, arg_order = sort_xr_nonxr_args(all_args)

        river_network = all_args["river_network"]
        return_type = all_args["return_type"]
        return_type = river_network.return_type if return_type is None else return_type
        return_grid = return_type == "gridded"

        if len(xr_args) == 0:
            output = func(**all_args)

            offset = 2 if return_grid else 1
            ndim = output.ndim
            dim_names = [f"axis{i + 1}" for i in range(ndim - offset)]
            coords = {
                dim: np.arange(size)
                for dim, size in zip(dim_names, output.shape[:-offset])
            }

            if return_grid:
                for k, v in river_network.coords.items():
                    coords[k] = v
                    dim_names.append(k)
            else:
                coords[node_default_coord] = np.arange(river_network.n_nodes)
                dim_names.append(node_default_coord)

            result = xr.DataArray(output, dims=dim_names, coords=coords, name="out")

            if not return_grid:
                coords_grid = np.meshgrid(*river_network.coords.values())
                assign_dict = {
                    k: (node_default_coord, v.flat[river_network.mask])
                    for k, v in zip(river_network.coords.keys(), coords_grid)
                }
                result = result.assign_coords(**assign_dict)
        else:

            reshuffled_func = get_reshuffled_func(func, arg_order)

            input_core_dims, output_core_dims = get_input_output_core_dims(
                input_core_dims, output_core_dims, xr_args, river_network, return_grid
            )

            output_sizes = (
                river_network.n_nodes
                if len(output_core_dims[0]) == 1
                else {k: v for k, v in zip(output_core_dims[0], river_network.shape)}
            )

            result = xr.apply_ufunc(
                reshuffled_func,
                *xr_args,
                input_core_dims=input_core_dims,
                output_core_dims=output_core_dims,
                dask_gufunc_kwargs={"output_sizes": output_sizes},
                output_dtypes=[float],
                dask="parallelized",
                kwargs=non_xr_kwargs,
            )

            if len(output_core_dims[0]) == 1:
                coords_grid = np.meshgrid(*river_network.coords.values())
                assign_dict = {
                    k: (output_core_dims[0], v.flat[river_network.mask])
                    for k, v in zip(river_network.coords.keys(), coords_grid)
                }
                result = result.assign_coords(**assign_dict)

        return result

    return wrapper



# from test_inputs.readers import *
# from test_inputs.subnetwork import *

# @pytest.mark.parametrize(
#     "river_network, mask, accumulate_downstream",
#     [
#         (("d8_ldd", d8_ldd_2), mask_2, masked_unit_accuflux_2),
#         (("cama_downxy", cama_downxy_2), mask_2, masked_unit_accuflux_2),
#         (("cama_nextxy", cama_nextxy_2), mask_2, masked_unit_accuflux_2),
#     ],
#     indirect=["river_network"],
# )
# def test_subnetwork(river_network, mask, accumulate_downstream):
#     river_network = river_network.create_subnetwork(mask)
#     field = np.ones(river_network.n_nodes)
#     accum = ekh.flow_downstream(river_network, field)
#     print(accum)
#     print(accumulate_downstream)
#     np.testing.assert_array_equal(accum, accumulate_downstream)



from ._accumulate import _ufunc_to_downstream
from .flow import propagate


def move_downstream(
    xp,
    river_network,
    field,
    func,
    node_additive_weight=None,
    node_multiplicative_weight=None,
    node_modifier_use_upstream=True,
    edge_additive_weight=None,
    edge_multiplicative_weight=None,
):
    invert_graph = False
    return move_python(
        xp,
        river_network,
        field,
        func,
        invert_graph,
        node_additive_weight,
        node_multiplicative_weight,
        node_modifier_use_upstream,
        edge_additive_weight,
        edge_multiplicative_weight,
    )


def move_upstream(
    xp,
    river_network,
    field,
    func,
    node_additive_weight=None,
    node_multiplicative_weight=None,
    node_modifier_use_upstream=True,
    edge_additive_weight=None,
    edge_multiplicative_weight=None,
):
    invert_graph = True
    return move_python(
        xp,
        river_network,
        field,
        func,
        invert_graph,
        node_additive_weight,
        node_multiplicative_weight,
        node_modifier_use_upstream,
        edge_additive_weight,
        edge_multiplicative_weight,
    )


def move_python(
    xp,
    river_network,
    field,
    func,
    invert_graph=False,
    node_additive_weight=None,
    node_multiplicative_weight=None,
    node_modifier_use_upstream=True,
    edge_additive_weight=None,
    edge_multiplicative_weight=None,
):
    op = _ufunc_to_downstream

    def operation(
        field,
        did,
        uid,
        eid,
        node_additive_weight,
        node_multiplicative_weight,
        node_modifier_use_upstream,
        edge_additive_weight,
        edge_multiplicative_weight,
    ):
        return op(
            field,
            did,
            uid,
            eid,
            node_additive_weight,
            node_multiplicative_weight,
            node_modifier_use_upstream,
            edge_additive_weight,
            edge_multiplicative_weight,
            func=func,
            xp=xp,
        )

    field = propagate(
        river_network,
        river_network.data,
        field,
        invert_graph,
        operation,
        node_additive_weight,
        node_multiplicative_weight,
        node_modifier_use_upstream,
        edge_additive_weight,
        edge_multiplicative_weight,
    )

    return field



from earthkit.hydro._core.online import calculate_online_metric
from earthkit.hydro._utils.decorators import mask, multi_backend


def calculate_downstream_metric(
    xp,
    river_network,
    field,
    metric,
    node_weights,
    edge_weights,
):
    return calculate_online_metric(
        xp,
        river_network,
        field,
        metric,
        node_weights,
        edge_weights,
        flow_direction="up",
    )


@multi_backend()
def var(xp, river_network, field, node_weights, edge_weights, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_calculate_downstream_metric = mask(return_type == "gridded")(
        calculate_downstream_metric
    )
    return decorated_calculate_downstream_metric(
        xp,
        river_network,
        field,
        "var",
        node_weights,
        edge_weights,
    )


@multi_backend()
def std(xp, river_network, field, node_weights, edge_weights, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_calculate_downstream_metric = mask(return_type == "gridded")(
        calculate_downstream_metric
    )
    return decorated_calculate_downstream_metric(
        xp,
        river_network,
        field,
        "std",
        node_weights,
        edge_weights,
    )


@multi_backend()
def mean(xp, river_network, field, node_weights, edge_weights, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_calculate_downstream_metric = mask(return_type == "gridded")(
        calculate_downstream_metric
    )
    return decorated_calculate_downstream_metric(
        xp,
        river_network,
        field,
        "mean",
        node_weights,
        edge_weights,
    )


@multi_backend()
def sum(xp, river_network, field, node_weights, edge_weights, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_calculate_downstream_metric = mask(return_type == "gridded")(
        calculate_downstream_metric
    )
    return decorated_calculate_downstream_metric(
        xp,
        river_network,
        field,
        "sum",
        node_weights,
        edge_weights,
    )


@multi_backend()
def min(xp, river_network, field, node_weights, edge_weights, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_calculate_downstream_metric = mask(return_type == "gridded")(
        calculate_downstream_metric
    )
    return decorated_calculate_downstream_metric(
        xp,
        river_network,
        field,
        "min",
        node_weights,
        edge_weights,
    )


@multi_backend()
def max(xp, river_network, field, node_weights, edge_weights, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_calculate_downstream_metric = mask(return_type == "gridded")(
        calculate_downstream_metric
    )
    return decorated_calculate_downstream_metric(
        xp,
        river_network,
        field,
        "max",
        node_weights,
        edge_weights,
    )



import os
import re

AUTODOCS_DIR = "source/autodocs"

for root, dirs, files in os.walk(AUTODOCS_DIR):
    for fname in files:
        if not fname.endswith(".rst"):
            continue
        path = os.path.join(root, fname)
        with open(path, "r", encoding="utf-8") as f:
            lines = f.readlines()

        new_lines = []
        changed = False
        i = 0

        while i < len(lines):
            line = lines[i].strip()

            if line == "earthkit.hydro package" and i + 1 < len(lines):
                new_title = "API Reference"
                underline = "=" * len(new_title)
                new_lines.append(f"{new_title}\n")
                new_lines.append(f"{underline}\n")
                i += 2  # skip underline
                changed = True
                continue

            # === Match lines like "earthkit.hydro.data\_structures package" ===
            match = re.match(r"^([\w\.\\]+)\s+(package|module)$", line)
            if match and i + 1 < len(lines):
                title = match.group(1)  # Keep escaped underscores as-is
                underline = "=" * len(title)
                new_lines.append(f"{title}\n")
                new_lines.append(f"{underline}\n")
                i += 2  # skip underline
                changed = True
                continue

            # === Add :titlesonly: under .. toctree:: if missing ===
            if line == ".. toctree::":
                new_lines.append(lines[i])  # original line
                i += 1
                # Look ahead for options
                has_titlesonly = False
                temp_lines = []
                while i < len(lines) and lines[i].lstrip().startswith(":"):
                    if ":titlesonly:" in lines[i]:
                        has_titlesonly = True
                    temp_lines.append(lines[i])
                    i += 1
                if not has_titlesonly:
                    new_lines.append("   :titlesonly:\n")
                    changed = True
                new_lines.extend(temp_lines)
                continue

            # default case
            new_lines.append(lines[i])
            i += 1

        if changed:
            with open(path, "w", encoding="utf-8") as f:
                f.writelines(new_lines)



from earthkit.hydro._core._find import _flow_find
from earthkit.hydro._utils.decorators import mask
from earthkit.hydro.upstream.array._operations import calculate_upstream_metric


def calculate_catchment_metric(
    xp,
    river_network,
    field,
    stations_1d,
    metric,
    node_weights,
    edge_weights,
):
    upstream_metric_field = calculate_upstream_metric(
        xp,
        river_network,
        field,
        metric,
        node_weights,
        edge_weights,
    )
    return xp.gather(upstream_metric_field, stations_1d, axis=-1)


@mask(unmask=False)
def var(xp, river_network, field, locations, node_weights, edge_weights):
    return calculate_catchment_metric(
        xp,
        river_network,
        field,
        locations,
        "var",
        node_weights,
        edge_weights,
    )


@mask(unmask=False)
def std(xp, river_network, field, locations, node_weights, edge_weights):
    return calculate_catchment_metric(
        xp,
        river_network,
        field,
        locations,
        "std",
        node_weights,
        edge_weights,
    )


@mask(unmask=False)
def mean(xp, river_network, field, locations, node_weights, edge_weights):
    return calculate_catchment_metric(
        xp,
        river_network,
        field,
        locations,
        "mean",
        node_weights,
        edge_weights,
    )


@mask(unmask=False)
def sum(xp, river_network, field, locations, node_weights, edge_weights):
    return calculate_catchment_metric(
        xp,
        river_network,
        field,
        locations,
        "sum",
        node_weights,
        edge_weights,
    )


@mask(unmask=False)
def min(xp, river_network, field, locations, node_weights, edge_weights):
    return calculate_catchment_metric(
        xp,
        river_network,
        field,
        locations,
        "min",
        node_weights,
        edge_weights,
    )


@mask(unmask=False)
def max(xp, river_network, field, locations, node_weights, edge_weights):
    return calculate_catchment_metric(
        xp,
        river_network,
        field,
        locations,
        "max",
        node_weights,
        edge_weights,
    )


def find(xp, river_network, field, overwrite, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_flow_find = mask(return_type == "gridded")(_flow_find)
    return decorated_flow_find(xp, river_network, field, overwrite)



from earthkit.hydro._core.move import calculate_move_metric


def upstream(xp, river_network, field, node_weights, edge_weights, metric):
    return calculate_move_metric(
        xp,
        river_network,
        field,
        metric,
        node_weights,
        edge_weights,
        flow_direction="up",
    )


def downstream(xp, river_network, field, node_weights, edge_weights, metric):
    return calculate_move_metric(
        xp,
        river_network,
        field,
        metric,
        node_weights,
        edge_weights,
        flow_direction="down",
    )



class RiverNetworkStorage:
    def __init__(
        self,
        n_nodes,
        n_edges,
        sorted_data,  # np.vstack((down_ids_upsort, up_ids_upsort, edge_ids_upsort))
        sources,
        sinks,
        coords,
        splits,  # indices of where to split sorted_data
        area,
        mask,
        shape,
        bifurcates=False,
        edge_weights=None,
    ):
        self.n_nodes = n_nodes
        self.n_edges = n_edges
        self.bifurcates = bifurcates
        self.sources = sources
        self.sinks = sinks
        self.coords = coords
        self.area = area
        self.sorted_data = sorted_data
        self.splits = splits
        self.mask = mask
        self.shape = shape
        self.edge_weights = edge_weights
        assert not (bifurcates and edge_weights is None)



from earthkit.hydro.move.array import _operations


def upstream(
    river_network,
    field,
    node_weights=None,
    edge_weights=None,
    metric="sum",
    return_type=None,
):
    r"""
    Moves a field upstream.

    Computes a one-step neighbor accumulation (local aggregation) moving upstream only.

    The accumulation is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        n_j &= \bigoplus_{i \in \mathrm{Down}(j)} w_{ij} \cdot x'_i
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Down}(j)` is the set of downstream nodes flowing out of node :math:`j`,
    - :math:`\bigoplus` is the aggregation function (e.g. a summation).
    - :math:`n_j` is the weighted aggregated value at node :math:`j`.

    Sinks are given a value of 0.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    metric : str, optional
        Aggregation function to apply. Options are 'var', 'std', 'mean', 'sum', 'min' and 'max'. Default is `'sum'`.
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.


    Returns
    -------
    array-like
        Array of values after movement up the river network for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.upstream(
        river_network, field, node_weights, edge_weights, metric, return_type
    )


def downstream(
    river_network,
    field,
    node_weights=None,
    edge_weights=None,
    metric="sum",
    return_type=None,
):
    r"""
    Moves a field downstream.

    Computes a one-step neighbor accumulation (local aggregation) moving downstream only.

    The accumulation is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        x'_i &= w'_i \cdot x_i \\
        n_j &= \bigoplus_{i \in \mathrm{Up}(j)} w_{ij} \cdot x'_i
        \end{align*}

    where:

    - :math:`x_i` is the input value at node :math:`i` (e.g., rainfall),
    - :math:`w'_i` is the node weight (e.g., pixel area),
    - :math:`w_{ij}` is the edge weight from node :math:`i` to node :math:`j` (e.g. discharge partitioning ratio),
    - :math:`\mathrm{Up}(j)` is the set of upstream nodes flowing into node :math:`j`,
    - :math:`\bigoplus` is the aggregation function (e.g. a summation).
    - :math:`n_j` is the weighted aggregated value at node :math:`j`.

    Sources are given a value of 0.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like
        An array containing field values defined on river network nodes or gridcells.
    node_weights : array-like, optional
        Array of weights for each river network node or gridcell. Default is None (unweighted).
    edge_weights : array-like, optional
        Array of weights for each edge. Default is None (unweighted).
    metric : str, optional
        Aggregation function to apply. Options are 'var', 'std', 'mean', 'sum', 'min' and 'max'. Default is `'sum'`.
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.


    Returns
    -------
    array-like
        Array of values after movement down the river network for every river network node or gridcell, depending on `return_type`.
    """
    return _operations.downstream(
        river_network, field, node_weights, edge_weights, metric, return_type
    )



from functools import wraps

import numpy as np
import xarray as xr

from earthkit.hydro._backends.find import get_array_backend
from earthkit.hydro._utils.coords import get_core_dims, node_default_coord
from earthkit.hydro._utils.decorators.xarray import (
    assert_xr_compatible_backend,
    get_full_signature,
    get_reshuffled_func,
    sort_xr_nonxr_args,
)
from earthkit.hydro._utils.locations import locations_to_1d


def get_input_core_dims(input_core_dims, xr_args):
    if input_core_dims is None:
        input_core_dims = [get_core_dims(xr_arg) for xr_arg in xr_args]
    elif len(input_core_dims) == 1:
        input_core_dims *= len(xr_args)

    return input_core_dims


def xarray(func):

    @wraps(func)
    def wrapper(*args, **kwargs):

        # Inspect the function signature and bind all arguments
        all_args = get_full_signature(func, *args, **kwargs)

        input_core_dims = all_args.pop("input_core_dims", None)

        assert_xr_compatible_backend(all_args["river_network"])

        river_network = all_args["river_network"]

        xp = get_array_backend(river_network.array_backend)

        locations = all_args["locations"]

        stations_1d, locations, orig_locations = locations_to_1d(
            xp, river_network, locations
        )

        all_args["locations"] = stations_1d

        # Separate xarray and non-xarray arguments
        xr_args, non_xr_kwargs, arg_order = sort_xr_nonxr_args(all_args)

        if len(xr_args) == 0:
            output = func(**all_args)

            ndim = output.ndim
            dim_names = [f"axis{i + 1}" for i in range(ndim - 1)]
            coords = {
                dim: np.arange(size) for dim, size in zip(dim_names, output.shape[:-1])
            }

            coords[node_default_coord] = np.arange(river_network.n_nodes)[stations_1d]
            dim_names.append(node_default_coord)

            result = xr.DataArray(output, dims=dim_names, coords=coords, name="out")

        else:

            reshuffled_func = get_reshuffled_func(func, arg_order)

            input_core_dims = get_input_core_dims(input_core_dims, xr_args)

            result = xr.apply_ufunc(
                reshuffled_func,
                *xr_args,
                input_core_dims=input_core_dims,
                output_core_dims=[[node_default_coord]],
                dask_gufunc_kwargs={"output_sizes": stations_1d.shape[0]},
                output_dtypes=[float],
                dask="parallelized",
                kwargs=non_xr_kwargs,
            )
            assign_dict = {
                node_default_coord: (
                    node_default_coord,
                    np.arange(river_network.n_nodes)[stations_1d],
                )
            }
            result = result.assign_coords(**assign_dict)

        coord1, coord2 = river_network.coords.values()
        coords_grid = np.meshgrid(coord2, coord1)
        assign_dict = {
            k: (node_default_coord, v.flat[river_network.mask][stations_1d])
            for k, v in zip(river_network.coords.keys(), coords_grid)
        }
        if isinstance(orig_locations, dict):
            assign_dict["name"] = (node_default_coord, list(orig_locations.keys()))
        result = result.assign_coords(**assign_dict)

        return result

    return wrapper



from functools import wraps


def mask(unmask=True):

    def decorator(func):

        @wraps(func)
        def wrapper(xp, river_network, field, *args, **kwargs):

            if field.shape[-2:] == river_network.shape:
                args, kwargs = process_args_kwargs(xp, river_network, args, kwargs)
                field_1d = mask_last2_dims(xp, field, river_network.mask, field.shape)

                out_1d = func(xp, river_network, field_1d, *args, **kwargs)

                if unmask:
                    out_shape = field.shape
                    return scatter_and_reshape(
                        xp, river_network.mask, out_1d, out_shape
                    )
                else:
                    return out_1d
            else:
                out_1d = func(xp, river_network, field, *args, **kwargs)
                if unmask:
                    out_shape = field.shape[:-2] + river_network.shape
                    return scatter_and_reshape(
                        xp, river_network.mask, out_1d, out_shape
                    )
                else:
                    return out_1d

        return wrapper

    return decorator


def mask_last2_dims(xp, tensor, mask, target_shape):
    B = target_shape[:-2]
    M, N = target_shape[-2], target_shape[-1]
    flat_shape = B + (M * N,)
    tensor_flat = xp.reshape(tensor, flat_shape)
    return xp.gather(tensor_flat, mask, axis=-1)


def scatter_and_reshape(xp, mask, out_1d, target_shape):
    B = target_shape[:-2]
    M, N = target_shape[-2], target_shape[-1]
    flat_shape = B + (M * N,)
    out_flat = xp.full(flat_shape, xp.nan)
    out_flat = xp.scatter_assign(out_flat, mask, out_1d)
    return xp.reshape(out_flat, target_shape)


def process_args_kwargs(xp, river_network, args, kwargs):
    def process_arg(arg):
        if (
            hasattr(arg, "shape")  # TODO: decide if robust enough
            and len(arg.shape) >= 2
            and arg.shape[-2:] == river_network.shape
        ):
            return mask_last2_dims(xp, arg, river_network.mask, arg.shape)
        return arg

    new_args = tuple(process_arg(arg) for arg in args)
    new_kwargs = {k: process_arg(v) for k, v in kwargs.items()}
    return new_args, new_kwargs



import earthkit.hydro.length.array as array
from earthkit.hydro._utils.decorators import xarray


@xarray
def min(
    river_network,
    locations,
    field=None,
    upstream=False,
    downstream=True,
    return_type=None,
    input_core_dims=None,
):
    r"""
    Calculates the minimum length to all points from a set of start
    locations.

    For each node in the network, calculates the minimum length starting from any of the start locations.

    The length is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        l_j &= w_j ~\text{for start locations}\\
        l_j &= \mathrm{min}(\infty,~\mathrm{min}_{i \in \mathrm{Neighbour}(j)} l_i) + w_j
        \end{align*}

    where:

    - :math:`w_i` is the node length (e.g., pixel length),
    - :math:`\mathrm{Neighbour}(j)` is the set of neighbouring nodes to node :math:`j`, which can include upstream and/or downstream nodes depending on passed arguments.
    - :math:`l_j` is the total length at node :math:`j`.

    Unreachable nodes are given a length of :math:`\infty`.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    locations : array-like or dict
        A list of source nodes.
    field : array-like or xarray object, optional
        An array containing length values defined on river network nodes or gridcells.
        Default is `xp.ones(river_network.n_nodes)`.
    upstream : bool, optional
        Whether or not to consider upstream lengths. Default is False.
    downstream : bool, optional
        Whether or not to consider downstream lengths. Default is True.
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.
    input_core_dims : sequence of sequence, optional
        List of core dimensions on each input xarray argument that should not be broadcast.
        Default is None, which attempts to autodetect input_core_dims from the xarray inputs.
        Ignored if no xarray inputs passed.

    Returns
    -------
    xarray object
        Array of minimum lengths for every river network node or gridcell, depending on `return_type`.
    """
    return array.min(river_network, locations, field, upstream, downstream, return_type)


@xarray
def max(
    river_network,
    locations,
    field=None,
    upstream=False,
    downstream=True,
    return_type=None,
    input_core_dims=None,
):
    r"""
    Calculates the maximum length to all points from a set of start
    locations.

    For each node in the network, calculates the maximum length starting from any of the start locations.

    The length is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        l_j &= w_j ~\text{for start locations}\\
        l_j &= \mathrm{max}(-\infty,~\mathrm{max}_{i \in \mathrm{Neighbour}(j)} l_i) + w_j
        \end{align*}

    where:

    - :math:`w_i` is the node length (e.g., pixel length),
    - :math:`\mathrm{Neighbour}(j)` is the set of neighbouring nodes to node :math:`j`, which can include upstream and/or downstream nodes depending on passed arguments.
    - :math:`l_j` is the total length at node :math:`j`.

    Unreachable nodes are given a length of :math:`-\infty`.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    locations : array-like or dict
        A list of source nodes.
    field : array-like or xarray object, optional
        An array containing length values defined on river network nodes or gridcells.
        Default is `xp.ones(river_network.n_nodes)`.
    upstream : bool, optional
        Whether or not to consider upstream lengths. Default is False.
    downstream : bool, optional
        Whether or not to consider downstream lengths. Default is True.
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.
    input_core_dims : sequence of sequence, optional
        List of core dimensions on each input xarray argument that should not be broadcast.
        Default is None, which attempts to autodetect input_core_dims from the xarray inputs.
        Ignored if no xarray inputs passed.

    Returns
    -------
    xarray object
        Array of maximum lengths for every river network node or gridcell, depending on `return_type`.
    """
    return array.max(river_network, locations, field, upstream, downstream, return_type)


@xarray
def to_source(
    river_network,
    field=None,
    path="shortest",
    return_type=None,
    input_core_dims=None,
):
    r"""
    Calculates the maximum length to all points from from the river network sources.

    For each node in the network, calculates the maximum length starting from any source.

    The length is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        l_j &= w_j ~\text{for sources}\\
        l_j &= \bigoplus \left(-\infty,~\bigoplus_{i \in \mathrm{Neighbour}(j)} l_i \right) + w_j
        \end{align*}

    where:

    - :math:`w_i` is the node length (e.g., pixel length),
    - :math:`\mathrm{Neighbour}(j)` is the set of neighbouring nodes to node :math:`j`, which can include upstream and/or downstream nodes depending on passed arguments.
    - :math:`\bigoplus` is the aggregation function (max for longest path or min for shortest path).
    - :math:`l_j` is the total length at node :math:`j`.

    Unreachable nodes are given a distance of :math:`-\infty` if :math:`\bigoplus` is a maximum, and :math:`\infty` if :math:`\bigoplus` is a minimum.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like or xarray object, optional
        An array containing length values defined on river network nodes or gridcells.
        Default is `xp.ones(river_network.n_nodes)`.
    path : str, optional
        Whether to compute the longest or shortest path. Default is "shortest".
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.
    input_core_dims : sequence of sequence, optional
        List of core dimensions on each input xarray argument that should not be broadcast.
        Default is None, which attempts to autodetect input_core_dims from the xarray inputs.
        Ignored if no xarray inputs passed.

    Returns
    -------
    xarray object
        Array of maximum lengths for every river network node or gridcell, depending on `return_type`.
    """
    return array.to_source(river_network, field, path, return_type)


@xarray
def to_sink(
    river_network,
    field=None,
    path="shortest",
    return_type=None,
    input_core_dims=None,
):
    r"""
    Calculates the maximum length to all points from from the river network sinks.

    For each node in the network, calculates the maximum length starting from any sink.

    The length is defined as:

    .. math::
        :nowrap:

        \begin{align*}
        l_j &= w_j ~\text{for sinks}\\
        l_j &= \bigoplus \left(-\infty,~\bigoplus_{i \in \mathrm{Neighbour}(j)} l_i \right) + w_j
        \end{align*}

    where:

    - :math:`w_i` is the node length (e.g., pixel length),
    - :math:`\mathrm{Neighbour}(j)` is the set of neighbouring nodes to node :math:`j`, which can include upstream and/or downstream nodes depending on passed arguments.
    - :math:`\bigoplus` is the aggregation function (max for longest path or min for shortest path).
    - :math:`l_j` is the total length at node :math:`j`.

    Unreachable nodes are given a distance of :math:`-\infty` if :math:`\bigoplus` is a maximum, and :math:`\infty` if :math:`\bigoplus` is a minimum.

    Parameters
    ----------
    river_network : RiverNetwork
        A river network object.
    field : array-like or xarray object, optional
        An array containing length values defined on river network nodes or gridcells.
        Default is `xp.ones(river_network.n_nodes)`.
    path : str, optional
        Whether to compute the longest or shortest path. Default is "shortest".
    return_type : str, optional
        Either "masked", "gridded" or None. If None (default), uses `river_network.return_type`.
    input_core_dims : sequence of sequence, optional
        List of core dimensions on each input xarray argument that should not be broadcast.
        Default is None, which attempts to autodetect input_core_dims from the xarray inputs.
        Ignored if no xarray inputs passed.

    Returns
    -------
    xarray object
        Array of maximum lengths for every river network node or gridcell, depending on `return_type`.
    """
    return array.to_sink(river_network, field, path, return_type)



import numpy as np
import pytest
from test_inputs.accumulation import *
from test_inputs.readers import *
from utils import convert_to_2d

import earthkit.hydro as ekh


def get_numpy_function(function_name):
    return getattr(np, function_name)


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1a,
        #     upstream_metric_sum_1a,
        #     mv_1a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1b,
        #     upstream_metric_sum_1b,
        #     mv_1b,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_sum_1c,
            mv_1c,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1d,
        #     upstream_metric_sum_1d,
        #     mv_1d,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_sum_1e,
            mv_1e,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1f,
        #     upstream_metric_sum_1f,
        #     mv_1f,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1g,
        #     upstream_metric_sum_1g,
        #     mv_1g,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2a,
        #     upstream_metric_sum_2a,
        #     mv_2a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2b,
        #     upstream_metric_sum_2b,
        #     mv_2b,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2g,
        #     upstream_metric_sum_2g,
        #     mv_2g,
        # ),
    ],
    indirect=["river_network"],
)
def test_upstream_metric_sum(river_network, input_field, flow_downstream, mv):
    output_field = ekh.upstream.array.sum(
        river_network, input_field, node_weights=None, return_type="masked"
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)

    print(input_field)
    input_field = convert_to_2d(river_network, input_field, 0)
    flow_downstream = convert_to_2d(river_network, flow_downstream, 0)
    print(mv, input_field.dtype)
    print(input_field, flow_downstream)
    output_field = ekh.upstream.array.sum(
        river_network, input_field, node_weights=None
    ).flatten()
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1a,
        #     upstream_metric_max_1a,
        #     mv_1a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1b,
        #     upstream_metric_max_1b,
        #     mv_1b,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_max_1c,
            mv_1c,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1d,
        #     upstream_metric_max_1d,
        #     mv_1d,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_max_1e,
            mv_1e,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1f,
        #     upstream_metric_max_1f,
        #     mv_1f,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1g,
        #     upstream_metric_max_1g,
        #     mv_1g,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2a,
        #     upstream_metric_max_2a,
        #     mv_2a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2b,
        #     upstream_metric_max_2b,
        #     mv_2b,
        # ),
    ],
    indirect=["river_network"],
)
def test_calculate_upstream_metric_max(river_network, input_field, flow_downstream, mv):
    output_field = ekh.upstream.array.max(
        river_network, input_field, node_weights=None, return_type="masked"
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1a,
        #     upstream_metric_min_1a,
        #     mv_1a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1b,
        #     upstream_metric_min_1b,
        #     mv_1b,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_min_1c,
            mv_1c,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1d,
        #     upstream_metric_min_1d,
        #     mv_1d,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_min_1e,
            mv_1e,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1f,
        #     upstream_metric_min_1f,
        #     mv_1f,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1g,
        #     upstream_metric_min_1g,
        #     mv_1g,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2a,
        #     upstream_metric_min_2a,
        #     mv_2a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2b,
        #     upstream_metric_min_2b,
        #     mv_2b,
        # ),
    ],
    indirect=["river_network"],
)
def test_calculate_upstream_metric_min(river_network, input_field, flow_downstream, mv):
    output_field = ekh.upstream.array.min(
        river_network, input_field, node_weights=None, return_type="masked"
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)


# # @pytest.mark.parametrize(
# #     "river_network, input_field, flow_downstream, mv",
# #     [
# #         (
# #             ("cama_nextxy", cama_nextxy_1),
# #             input_field_1a,
# #             upstream_metric_product_1a,
# #             mv_1a,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_1),
# #             input_field_1b,
# #             upstream_metric_product_1b,
# #             mv_1b,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_1),
# #             input_field_1c,
# #             upstream_metric_product_1c,
# #             mv_1c,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_1),
# #             input_field_1d,
# #             upstream_metric_product_1d,
# #             mv_1d,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_1),
# #             input_field_1e,
# #             upstream_metric_product_1e,
# #             mv_1e,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_1),
# #             input_field_1f,
# #             upstream_metric_product_1f,
# #             mv_1f,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_1),
# #             input_field_1g,
# #             upstream_metric_product_1g,
# #             mv_1g,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_2),
# #             input_field_2a,
# #             upstream_metric_product_2a,
# #             mv_2a,
# #         ),
# #         (
# #             ("cama_nextxy", cama_nextxy_2),
# #             input_field_2b,
# #             upstream_metric_product_2b,
# #             mv_2b,
# #         ),
# #     ],
# #     indirect=["river_network"],
# # )
# # def test_calculate_upstream_metric_prod(
# #     river_network, input_field, flow_downstream, mv
# # ):
# #     output_field = ekh.upstream.array.prod(
# #         river_network,
# #         input_field,
# #         node_weights=None,
# #         mv=mv,
# #         accept_missing=True,
# #     )
# #     print(output_field)
# #     print(flow_downstream)
# #     assert output_field.dtype == flow_downstream.dtype
# #     np.testing.assert_allclose(output_field, flow_downstream)


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1a,
        #     upstream_metric_mean_1a,
        #     mv_1a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1b,
        #     upstream_metric_mean_1b,
        #     mv_1b,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_mean_1c,
            mv_1c,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1d,
        #     upstream_metric_mean_1d,
        #     mv_1d,
        # ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_mean_1e,
            mv_1e,
        ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1f,
        #     upstream_metric_mean_1f,
        #     mv_1f,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_1),
        #     input_field_1g,
        #     upstream_metric_mean_1g,
        #     mv_1g,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2a,
        #     upstream_metric_mean_2a,
        #     mv_2a,
        # ),
        # (
        #     ("cama_nextxy", cama_nextxy_2),
        #     input_field_2b,
        #     upstream_metric_mean_2b,
        #     mv_2b,
        # ),
    ],
    indirect=["river_network"],
)
def test_calculate_upstream_metric_mean(
    river_network, input_field, flow_downstream, mv
):
    output_field = ekh.upstream.array.mean(
        river_network, input_field, node_weights=None, return_type="masked"
    )
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)

    input_field = convert_to_2d(river_network, input_field, 0)
    flow_downstream = convert_to_2d(river_network, flow_downstream, 0)
    output_field = ekh.upstream.array.mean(
        river_network,
        input_field,
        node_weights=None,
    ).flatten()
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)


# @pytest.mark.parametrize(
#     "river_network, input_field, accum_field",
#     [
#         (("d8_ldd", d8_ldd_1), input_field_1b, upstream_metric_sum_1g),
#         (("cama_downxy", cama_downxy_1), input_field_1b, upstream_metric_sum_1g),
#         (("cama_nextxy", cama_nextxy_1), input_field_1b, upstream_metric_sum_1g),
#         (("d8_ldd", d8_ldd_2), input_field_2b, upstream_metric_sum_2g),
#         (("cama_downxy", cama_downxy_2), input_field_2b, upstream_metric_sum_2g),
#         (("cama_nextxy", cama_nextxy_2), input_field_2b, upstream_metric_sum_2g),
#     ],
#     indirect=["river_network"],
# )
# def test_accumulate_downstream_missing(river_network, input_field, accum_field):
#     accum = ekh.upstream.array.sum(river_network, input_field, mv=-1, accept_missing=True)
#     print(accum)
#     print(accum_field)
#     np.testing.assert_array_equal(accum, accum_field)


# @pytest.mark.parametrize(
#     "river_network",
#     [
#         ("d8_ldd", d8_ldd_1),
#         ("cama_downxy", cama_downxy_1),
#         ("cama_nextxy", cama_nextxy_1),
#         ("d8_ldd", d8_ldd_2),
#         ("cama_downxy", cama_downxy_2),
#         ("cama_nextxy", cama_nextxy_2),
#     ],
#     indirect=True,
# )
# @pytest.mark.parametrize("N", range(4))
# def test_accumulate_downstream_2d(river_network, N):
#     field = np.random.rand(*([np.random.randint(10)] * N), *river_network.mask.shape)
#     field_1d = field[..., river_network.mask]
#     accum = ekh.upstream.array.sum(river_network, field_1d)
#     np.testing.assert_array_equal(
#         accum, ekh.upstream.array.sum(river_network, field)[..., river_network.mask]
#     )
#     np.testing.assert_array_equal(
#         ekh.upstream.array.sum(river_network, field)[..., ~river_network.mask],
#         field[..., ~river_network.mask],
#     )


# # @pytest.mark.parametrize(
# #     "river_network",
# #     [
# #         ("d8_ldd", d8_ldd_1),
# #         ("cama_downxy", cama_downxy_1),
# #         ("cama_nextxy", cama_nextxy_1),
# #         ("d8_ldd", d8_ldd_2),
# #         ("cama_downxy", cama_downxy_2),
# #         ("cama_nextxy", cama_nextxy_2),
# #     ],
# #     indirect=True,
# # )
# # @pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "std", "var"])
# # def test_calculate_catchment(river_network, metric):
# #     from earthkit.hydro.catchments.metrics import calculate_catchment_metric
# #     from earthkit.hydro.upstream.operations import calculate_upstream_metric

# #     field = np.random.rand(3, 4, river_network.n_nodes)
# #     node_weights = np.random.rand(3, 4, river_network.n_nodes)
# #     catchment_metric = calculate_catchment_metric(
# #         river_network, field, river_network.nodes, metric, node_weights
# #     )
# #     upstream_field = calculate_upstream_metric(
# #         river_network, field, metric, node_weights
# #     )
# #     print(catchment_metric[0].shape, upstream_field.shape)
# #     for i in catchment_metric.keys():
# #         assert catchment_metric[i].dtype == upstream_field.dtype
# #         np.testing.assert_allclose(catchment_metric[i], upstream_field[..., i])


# # @pytest.mark.parametrize(
# #     "river_network",
# #     [
# #         ("d8_ldd", d8_ldd_1),
# #         ("cama_downxy", cama_downxy_1),
# #         ("cama_nextxy", cama_nextxy_1),
# #         ("d8_ldd", d8_ldd_2),
# #         ("cama_downxy", cama_downxy_2),
# #         ("cama_nextxy", cama_nextxy_2),
# #     ],
# #     indirect=True,
# # )
# # @pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "prod", "std", "var"])
# # def test_calculate_subcatchment(river_network, metric):
# #     from earthkit.hydro.catchments.metrics import calculate_catchment_metric
# #     from earthkit.hydro.subcatchments.metrics import calculate_subcatchment_metric

# #     shape = (3, 4)
# #     field = np.random.rand(*shape, river_network.n_nodes) + 1
# #     node_weights = np.random.rand(*shape, river_network.n_nodes) + 1
# #     subcatchment_metric = calculate_subcatchment_metric(
# #         river_network, field, river_network.sinks, metric, node_weights
# #     )
# #     catchment_metric = calculate_catchment_metric(
# #         river_network, field, river_network.sinks, metric, node_weights
# #     )
# #     for i in catchment_metric.keys():
# #         assert catchment_metric[i].dtype == subcatchment_metric[i].dtype
# #         # the two methods don't match always because of numerical instabilities in the
# #         # variance computation arising from the different scales
# #         np.testing.assert_allclose(
# #             subcatchment_metric[i], catchment_metric[i], atol=5e-8
# #         )

# #     subcatchment_metric = calculate_subcatchment_metric(
# #         river_network, field, river_network.nodes, metric, node_weights
# #     )
# #     catchment_metric = calculate_catchment_metric(
# #         river_network,
# #         field,
# #         np.array(
# #             [
# #                 i
# #                 for i in river_network.nodes
# #                 if i not in river_network.sinks and i not in river_network.sources
# #             ]
# #         ),
# #         metric,
# #         node_weights,
# #     )
# #     for i in catchment_metric.keys():
# #         assert catchment_metric[i].dtype == subcatchment_metric[i].dtype
# #         np.testing.assert_raises(
# #             AssertionError,
# #             np.testing.assert_allclose,
# #             subcatchment_metric[i],
# #             catchment_metric[i],
# #         )


# # @pytest.mark.parametrize(
# #     "river_network",
# #     [
# #         ("d8_ldd", d8_ldd_1),
# #         ("cama_downxy", cama_downxy_1),
# #         ("cama_nextxy", cama_nextxy_1),
# #         ("d8_ldd", d8_ldd_2),
# #         ("cama_downxy", cama_downxy_2),
# #         ("cama_nextxy", cama_nextxy_2),
# #     ],
# #     indirect=True,
# # )
# # @pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "prod", "std", "var"])
# # def test_calculate_catchment_numpy(river_network, metric):
# #     from earthkit.hydro.catchments.metrics import calculate_catchment_metric

# #     shape = (3, 4)
# #     field = np.random.rand(*shape, river_network.n_nodes)
# #     node_weights = None
# #     nodes = np.random.choice(river_network.n_nodes, 5)
# #     catchment_metric = calculate_catchment_metric(
# #         river_network, field, nodes, metric, node_weights
# #     )
# #     numpy_func = get_numpy_function(metric)
# #     for i in catchment_metric.keys():
# #         mask = np.zeros(river_network.n_nodes, dtype=bool)
# #         mask[i] = True
# #         mask = ekh.catchments.find(river_network, mask)
# #         numpy_answer = numpy_func(field[..., mask], axis=-1)

# #         np.testing.assert_allclose(catchment_metric[i], numpy_answer)


# # @pytest.mark.parametrize(
# #     "river_network",
# #     [
# #         ("d8_ldd", d8_ldd_1),
# #         ("cama_downxy", cama_downxy_1),
# #         ("cama_nextxy", cama_nextxy_1),
# #         ("d8_ldd", d8_ldd_2),
# #         ("cama_downxy", cama_downxy_2),
# #         ("cama_nextxy", cama_nextxy_2),
# #     ],
# #     indirect=True,
# # )
# # @pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "prod", "std", "var"])
# # def test_calculate_subcatchment_numpy(river_network, metric):
# #     from earthkit.hydro.subcatchments.metrics import calculate_subcatchment_metric

# #     shape = (3, 4)
# #     field = np.random.rand(*shape, river_network.n_nodes)
# #     node_weights = None
# #     nodes = np.random.choice(river_network.n_nodes, 5, replace=False)
# #     subcatchment_metric = calculate_subcatchment_metric(
# #         river_network, field, nodes, metric, node_weights
# #     )
# #     numpy_func = get_numpy_function(metric)
# #     labels = np.zeros(river_network.n_nodes, dtype=int)
# #     labels[nodes] = nodes + 1
# #     labels = ekh.subcatchments.find(river_network, labels)
# #     for i in subcatchment_metric.keys():
# #         mask = labels == (i + 1)
# #         numpy_answer = numpy_func(field[..., mask], axis=-1)

# #         np.testing.assert_allclose(subcatchment_metric[i], numpy_answer)



import jax
import jax.numpy as jnp

from .array_backend import ArrayBackend


class JAXBackend(ArrayBackend):
    def __init__(self):
        super().__init__(jnp)

    @property
    def name(self):
        return "jax"

    def copy(self, x):
        return x

    def gather(self, arr, indices, axis=-1):
        assert axis == -1
        return arr[..., indices]

    def scatter_assign(self, target, indices, updates):
        return target.at[..., indices].set(updates)

    def scatter_add(self, target, indices, updates):
        return target.at[(*[slice(None)] * (target.ndim - 1), indices)].add(updates)

    def asarray(self, arr, dtype=None, device=None, copy=None):
        for d in jax.devices():
            if d.platform == device:
                device = d
                break
        return jnp.asarray(arr, dtype=dtype, order=None, copy=copy, device=device)



def get_core_grid_dims(ds):
    possible_names = [["lat", "lon"], ["latitude", "longitude"], ["y", "x"]]
    for names in possible_names:
        present = True
        for name in names:
            present &= name in ds.coords
        if present:
            return names


def get_core_node_dims(ds):
    possible_names = [
        ["index"],
        ["node_index"],
        ["node_id"],
        ["station_index"],
        ["station_id"],
        ["gauge_id"],
        ["id"],
        ["idx"],
    ]
    for names in possible_names:
        present = True
        for name in names:
            present &= name in ds.coords
        if present:
            return names


def get_core_edge_dims(ds):
    possible_names = [["edge_id"]]
    for names in possible_names:
        present = True
        for name in names:
            present &= name in ds.coords
        if present:
            return names


def get_core_dims(ds):
    dims = get_core_grid_dims(ds)
    if dims is None:
        dims = get_core_node_dims(ds)
    if dims is None:
        dims = get_core_edge_dims(ds)
    if dims is None:
        raise ValueError("Could not autodetect xarray core dims.")
    return dims


node_default_coord = "node_index"



import os
import tempfile
from functools import wraps
from hashlib import sha256

import joblib

from earthkit.hydro._version import __version__ as ekh_version
from earthkit.hydro.data_structures._network import RiverNetwork

# read in only up to second decimal point
# i.e. 0.1.dev90 -> 0.1
ekh_version = ".".join(ekh_version.split(".")[:2])


def cache(func):
    """
    Decorator to allow automatic use of cache.

    Parameters
    ----------
    func : callable
        The function to be wrapped and executed with masking applied.

    Returns
    -------
    callable
        The wrapped function.
    """

    @wraps(func)
    def wrapper(
        path,
        river_network_format,
        source="file",
        use_cache=True,
        cache_dir=tempfile.mkdtemp(suffix="_earthkit_hydro"),
        cache_fname="{ekh_version}_{hash}.joblib",
        cache_compression=1,
    ):
        """
        Wrapper to load river network from cache if available, otherwise
        create and cache it.

        Parameters
        ----------
        path : str
            The path to the river network.
        river_network_format : str
            The format of the river network file.
            Supported formats are "precomputed", "cama", "pcr_d8", and "esri_d8".
        source : str, optional
            The source of the river network.
            For possible sources see:
            https://earthkit-data.readthedocs.io/en/latest/guide/sources.html
        use_cache : bool, optional
            Whether to use caching. Default is True.
        cache_dir : str, optional
            The directory to store the cache files. Default is a temporary directory.
        cache_fname : str, optional
            The filename template for the cache files.
            Default is "{ekh_version}_{hash}.joblib".
        cache_compression : int, optional
            The compression level for the cache files. Default is 1.

        Returns
        -------
        earthkit.hydro.network_class.RiverNetwork
            The loaded river network.
        """
        if use_cache:
            hashed_name = sha256(path.encode("utf-8")).hexdigest()
            cache_dir = cache_dir.format(ekh_version=ekh_version, hash=hashed_name)
            cache_fname = cache_fname.format(ekh_version=ekh_version, hash=hashed_name)
            cache_filepath = os.path.join(cache_dir, cache_fname)

            if os.path.isfile(cache_filepath):
                print(f"Loading river network from cache ({cache_filepath}).")
                return RiverNetwork(joblib.load(cache_filepath))
            else:
                print(f"River network not found in cache ({cache_filepath}).")
                os.makedirs(cache_dir, exist_ok=True)
        else:
            print("Cache disabled.")

        network = func(path, river_network_format, source)

        if use_cache:
            joblib.dump(network._storage, cache_filepath, compress=cache_compression)
            print(f"River network loaded, saving to cache ({cache_filepath}).")

        return network

    return wrapper



import tensorflow as tf

from .array_backend import ArrayBackend


class TFBackend(ArrayBackend):
    def __init__(self):
        super().__init__(tf)

    @property
    def name(self):
        return "tensorflow"

    def copy(self, x):
        return x

    def scatter_assign(self, target, indices, updates):
        target_shape = tf.shape(target)
        batch_dims = target_shape[:-1]
        num_batch = tf.reduce_prod(batch_dims)
        num_idx = tf.shape(indices)[0]

        flat_target = tf.reshape(target, (num_batch, -1))
        flat_values = tf.reshape(updates, (num_batch, num_idx))

        batch_range = tf.range(num_batch)[:, None]
        batch_ids = tf.tile(batch_range, [1, num_idx])
        scatter_idx = tf.stack(
            [batch_ids, tf.tile(tf.expand_dims(indices, 0), [num_batch, 1])], axis=-1
        )
        scatter_idx = tf.reshape(scatter_idx, (-1, 2))

        scatter_vals = tf.reshape(flat_values, (-1,))
        flat_result = tf.tensor_scatter_nd_update(
            flat_target, scatter_idx, scatter_vals
        )

        return tf.reshape(flat_result, target_shape)

    def scatter_add(self, target, indices, updates):
        target_shape = tf.shape(target)
        batch_shape = target_shape[:-1]
        depth = target_shape[-1]
        num_indices = tf.shape(indices)[0]

        flat_batch_size = tf.reduce_prod(batch_shape)
        K = num_indices
        D = depth
        updates_flat = tf.reshape(updates, [-1])

        segment_ids = tf.tile(indices, [flat_batch_size])

        batch_ids = tf.repeat(tf.range(flat_batch_size, dtype=tf.int32), repeats=K)

        combined_segments = batch_ids * D + segment_ids

        scattered_flat = tf.math.unsorted_segment_sum(
            data=updates_flat,
            segment_ids=combined_segments,
            num_segments=flat_batch_size * D,
        )

        scattered = tf.reshape(scattered_flat, [flat_batch_size, D])

        result = tf.reshape(scattered, tf.concat([batch_shape, [depth]], axis=0))
        return target + result

    def gather(self, arr, indices, axis=-1):
        return tf.gather(arr, indices, axis=axis)

    def full_like(self, arr, value):
        return tf.fill(tf.shape(arr), value)

    def full(self, shape, value):
        return tf.fill(shape, value)

    @property
    def nan(self):
        return float("nan")

    def asarray(self, arr, dtype=None, device=None, copy=None):
        tensor = tf.convert_to_tensor(arr, dtype=dtype)

        if copy and device is None:
            tensor = tf.identity(tensor)

        if device is not None:
            with tf.device(device):
                tensor = tf.identity(tensor)
        return tensor



import array_api_compat.torch as torch

from .array_backend import ArrayBackend


class TorchBackend(ArrayBackend):
    def __init__(self):
        super().__init__(torch)

    @property
    def name(self):
        return "torch"

    def copy(self, x):
        return x.clone()

    def gather(self, arr, indices, axis=-1):
        return torch.index_select(arr, dim=axis, index=indices)

    def scatter_assign(self, target, indices, updates):
        target[..., indices] = updates
        return target

    def scatter_add(self, target, indices, updates):
        return target.index_add(-1, indices, updates)

    def scatter_max(self, target, indices, updates):
        return target.index_reduce(-1, indices, updates, "amax")

    def scatter_min(self, target, indices, updates):
        return target.index_reduce(-1, indices, updates, "amin")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

on_rtd = os.environ.get("READTHEDOCS") == "True"

if on_rtd:
    version = os.environ.get("READTHEDOCS_VERSION", "latest")
    release = version
else:
    version = "dev"
    release = "dev"

rtd_version = version if version != "latest" else "develop"
rtd_version_type = os.environ.get("READTHEDOCS_VERSION_TYPE", "branch")

if rtd_version_type in ("branch", "tag"):
    source_branch = rtd_version
else:
    source_branch = "main"

sys.path.insert(0, os.path.abspath("../../src"))

project = "earthkit-hydro"
copyright = "2025, European Centre for Medium-Range Weather Forecasts (ECMWF)"
author = "European Centre for Medium-Range Weather Forecasts (ECMWF)"
# release = '0.0.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # Automatically extracts documentation from your Python docstrings
    "sphinx.ext.autodoc",
    # Supports Google-style and NumPy-style docstrings
    "sphinx.ext.napoleon",
    # Renders LaTeX math in HTML using MathJax
    "sphinx.ext.mathjax",
    # Option to click viewcode
    "sphinx.ext.viewcode",
    # Links to the documentation of other projects via cross-references
    # "sphinx.ext.intersphinx",
    # Generates summary tables for modules/classes/functions
    # "sphinx.ext.autosummary",
    # Allows citing BibTeX bibliographic entries in reStructuredText
    "sphinxcontrib.bibtex",
    # Tests snippets in documentation by running embedded Python examples
    # "sphinx.ext.doctest",
    # Checks documentation coverage of the codebase
    # "sphinx.ext.coverage",
    # Adds .nojekyll file and helps configure docs for GitHub Pages hosting
    # "sphinx.ext.githubpages",
    # Adds "Edit on GitHub" links to documentation pages
    # "edit_on_github",
    # Adds "Edit on GitHub" links to documentation pages
    # "sphinx_github_style",
    # Option to link to code
    # "sphinx.ext.linkcode",
    # Automatically includes type hints from function signatures into the documentation
    # "sphinx_autodoc_typehints",
    # Integrates Jupyter Notebooks into Sphinx
    "nbsphinx",
]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


html_theme = "furo"

html_static_path = ["_static"]

# html_context = {
#     "display_github": True,
#     "github_user": "ecmwf",         # GitHub username
#     "github_repo": "docsample",             # GitHub repository name
#     "github_version": "main",               # Branch (e.g., 'main', 'master')
#     "conf_py_path": "/docs/",               # Path to your docs root in the repo
# }

bibtex_bibfiles = ["references.bib"]

html_theme_options = {
    "light_css_variables": {
        "color-sidebar-background": "#001F3F",
        "color-sidebar-link-text": "#ffffff",
        "color-sidebar-brand-text": "#ffffff",
        "color-sidebar-caption-text": "#ffffff",
        "color-brand-primary": "#00D9FF",
        "color-brand-content": "#5f8dd3",
    },
    "dark_css_variables": {
        "color-sidebar-background": "#001F3F",
        "color-sidebar-link-text": "#ffffff",
        "color-sidebar-brand-text": "#ffffff",
        "color-sidebar-caption-text": "#ffffff",
        "color-brand-primary": "#00D9FF",
        "color-brand-content": "#5f8dd3",
    },
    "light_logo": "earthkit-hydro-dark.svg",
    "dark_logo": "earthkit-hydro-dark.svg",
    "source_repository": "https://github.com/ecmwf/earthkit-hydro/",
    "source_branch": source_branch,
    "source_directory": "docs/source",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/ecmwf/earthkit-hydro",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}



from earthkit.hydro._utils.decorators import multi_backend
from earthkit.hydro._utils.locations import locations_to_1d
from earthkit.hydro.catchments.array import __operations as _operations


@multi_backend(allow_jax_jit=False)
def var(xp, river_network, field, locations, node_weights, edge_weights):
    stations_1d, _, _ = locations_to_1d(xp, river_network, locations)
    return _operations.var(
        xp, river_network, field, stations_1d, node_weights, edge_weights
    )


@multi_backend(allow_jax_jit=False)
def std(xp, river_network, field, locations, node_weights, edge_weights):
    stations_1d, _, _ = locations_to_1d(xp, river_network, locations)
    return _operations.std(
        xp, river_network, field, stations_1d, node_weights, edge_weights
    )


@multi_backend(allow_jax_jit=False)
def mean(xp, river_network, field, locations, node_weights, edge_weights):
    stations_1d, _, _ = locations_to_1d(xp, river_network, locations)
    return _operations.mean(
        xp, river_network, field, stations_1d, node_weights, edge_weights
    )


@multi_backend(allow_jax_jit=False)
def sum(xp, river_network, field, locations, node_weights, edge_weights):
    stations_1d, _, _ = locations_to_1d(xp, river_network, locations)
    return _operations.sum(
        xp, river_network, field, stations_1d, node_weights, edge_weights
    )


@multi_backend(allow_jax_jit=False)
def min(xp, river_network, field, locations, node_weights, edge_weights):
    stations_1d, _, _ = locations_to_1d(xp, river_network, locations)
    return _operations.min(
        xp, river_network, field, stations_1d, node_weights, edge_weights
    )


@multi_backend(allow_jax_jit=False)
def max(xp, river_network, field, locations, node_weights, edge_weights):
    stations_1d, _, _ = locations_to_1d(xp, river_network, locations)
    return _operations.max(
        xp, river_network, field, stations_1d, node_weights, edge_weights
    )


@multi_backend()
def find(xp, river_network, locations, overwrite, return_type):
    stations1d, _, _ = locations_to_1d(xp, river_network, locations)
    field = xp.full(river_network.n_nodes, xp.nan, device=river_network.device)
    field[stations1d] = xp.arange(stations1d.shape[0])
    return _operations.find(xp, river_network, field, overwrite, return_type)



# import numpy as np
# import pytest
# from test_inputs.distance import *
# from test_inputs.readers import *

# import earthkit.hydro as ekh


# @pytest.mark.parametrize(
#     "river_network, stations_list, upstream, downstream, weights, result",
#     [
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             True,
#             True,
#             weights_1,
#             distance_1_min_up_down,
#         ),
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             True,
#             False,
#             weights_1,
#             distance_1_min_up,
#         ),
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             False,
#             True,
#             weights_1,
#             distance_1_min_down,
#         ),
#     ],
#     indirect=["river_network"],
# )
# def test_distance_min(
#     river_network, stations_list, upstream, downstream, weights, result
# ):
#     dist = ekh.distance.min(
#         river_network,
#         stations_list,
#         upstream=upstream,
#         downstream=downstream,
#         weights=weights,
#     )
#     np.testing.assert_allclose(dist, result)


# @pytest.mark.parametrize(
#     "river_network, stations_list, upstream, downstream, weights, result",
#     [
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             True,
#             False,
#             weights_1,
#             distance_1_max_up,
#         ),
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             False,
#             True,
#             weights_1,
#             distance_1_max_down,
#         ),
#     ],
#     indirect=["river_network"],
# )
# def test_distance_max(
#     river_network, stations_list, upstream, downstream, weights, result
# ):
#     dist = ekh.distance.max(
#         river_network,
#         stations_list,
#         upstream=upstream,
#         downstream=downstream,
#         weights=weights,
#     )
#     np.testing.assert_allclose(dist, result)


# @pytest.mark.parametrize(
#     "river_network, stations_list, upstream, downstream, weights, result",
#     [
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             True,
#             True,
#             weights_1,
#             length_1_min_up_down,
#         ),
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             True,
#             False,
#             weights_1,
#             length_1_min_up,
#         ),
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             False,
#             True,
#             weights_1,
#             length_1_min_down,
#         ),
#     ],
#     indirect=["river_network"],
# )
# def test_length_min(
#     river_network, stations_list, upstream, downstream, weights, result
# ):
#     length = ekh.length.min(
#         river_network,
#         stations_list,
#         upstream=upstream,
#         downstream=downstream,
#         weights=weights,
#     )
#     np.testing.assert_allclose(length, result)


# @pytest.mark.parametrize(
#     "river_network, stations_list, upstream, downstream, weights, result",
#     [
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             True,
#             False,
#             weights_1,
#             length_1_max_up,
#         ),
#         (
#             ("cama_nextxy", cama_nextxy_1),
#             stations,
#             False,
#             True,
#             weights_1,
#             length_1_max_down,
#         ),
#     ],
#     indirect=["river_network"],
# )
# def test_length_max(
#     river_network, stations_list, upstream, downstream, weights, result
# ):
#     length = ekh.length.max(
#         river_network,
#         stations_list,
#         upstream=upstream,
#         downstream=downstream,
#         weights=weights,
#     )
#     np.testing.assert_allclose(length, result)



import copy as cp

from earthkit.hydro._backends.numpy_backend import NumPyBackend
from earthkit.hydro._utils.decorators.masking import mask_last2_dims
from earthkit.hydro.data_structures import RiverNetwork

np = NumPyBackend()


def from_mask(river_network: RiverNetwork, node_mask=None, edge_mask=None, copy=True):
    """
    Create a subnetwork from a river network.

    Parameters
    ----------
    river_network : RiverNetwork
        Original river network from which to create a subnetwork.
    node_mask : array, optional
        A mask of the network nodes or gridcells. Default is None (all True).
    edge_mask : array, optional
        A mask of the network edges. Default is None (all True).
    copy : bool, optional
        Whether or not to modify the original river network or return a copy. Default is True.

    Returns
    -------
    RiverNetwork
        The river network object created from the given data.
    """
    if river_network.array_backend != "numpy" or copy is not True:
        raise NotImplementedError

    if node_mask is None and edge_mask is None:
        return cp.deepcopy(river_network)

    if node_mask is not None:
        if node_mask.shape[-2:] == river_network.shape:
            node_mask = mask_last2_dims(
                np, node_mask, river_network.mask, node_mask.shape
            )

    node_relabel = np.empty(river_network.n_nodes, dtype=int)
    node_relabel[node_mask] = np.arange(node_mask.sum())

    storage = cp.deepcopy(river_network._storage)
    if edge_mask is not None and node_mask is not None:
        valid_edges = edge_mask[storage.sorted_data[2]] & (
            node_mask[storage.sorted_data[0]] & node_mask[storage.sorted_data[1]]
        )
    elif edge_mask is None:
        valid_edges = (
            node_mask[storage.sorted_data[0]] & node_mask[storage.sorted_data[1]]
        )
    else:
        valid_edges = edge_mask[storage.sorted_data[2]]

    original_order_edge_mask = np.empty(river_network.n_edges, dtype=bool)
    original_order_edge_mask[storage.sorted_data[2]] = valid_edges
    edge_relabel = np.empty(river_network.n_edges, dtype=int)
    edge_relabel[original_order_edge_mask] = np.arange(original_order_edge_mask.sum())

    storage.sorted_data = storage.sorted_data[..., valid_edges]
    storage.sorted_data[0] = node_relabel[storage.sorted_data[0]]
    storage.sorted_data[1] = node_relabel[storage.sorted_data[1]]
    storage.sorted_data[2] = edge_relabel[storage.sorted_data[2]]

    storage.splits = np.cumsum(valid_edges)[storage.splits - 1]
    storage.mask = storage.mask[node_mask]
    storage.n_nodes = storage.mask.shape[0]
    storage.n_edges = storage.sorted_data.shape[1]

    return RiverNetwork(storage)


def crop(river_network, copy=True):
    """
    Crop a gridded network to the minimum bounding grid.

    Parameters
    ----------
    river_network : RiverNetwork
        Original river network from which to create a cropped network.
    copy : bool, optional
        Whether or not to modify the original river network or return a copy. Default is True.

    Returns
    -------
    RiverNetwork
        The river network object created from the given data.
    """

    if river_network.array_backend != "numpy" or copy is not True:
        raise NotImplementedError

    storage = cp.deepcopy(river_network._storage)

    rows, cols = np.unravel_index(storage.mask, shape=(storage.shape))

    row_min, row_max = rows.min(), rows.max()
    col_min, col_max = cols.min(), cols.max()

    storage.shape = (int(row_max - row_min + 1), int(col_max - col_min + 1))

    storage.mask = np.ravel_multi_index(
        (rows - row_min, cols - col_min), dims=storage.shape
    )

    for i, key in enumerate(storage.coords.keys()):
        if i == 0:
            storage.coords[key] = storage.coords[key][row_min : row_max + 1]
        elif i == 1:
            storage.coords[key] = storage.coords[key][col_min : col_max + 1]
        else:
            raise ValueError("coords must not have more than 2 keys.")

    return RiverNetwork(storage)



import earthkit.hydro.catchments.array.__operations as array
from earthkit.hydro._utils.decorators import multi_backend


@multi_backend(allow_jax_jit=False)
def var(
    xp,
    river_network,
    field,
    locations,
    node_weights,
    edge_weights,
):
    return array.var(xp, river_network, field, locations, node_weights, edge_weights)


@multi_backend(allow_jax_jit=False)
def std(
    xp,
    river_network,
    field,
    locations,
    node_weights,
    edge_weights,
):
    return array.std(xp, river_network, field, locations, node_weights, edge_weights)


@multi_backend(allow_jax_jit=False)
def mean(
    xp,
    river_network,
    field,
    locations,
    node_weights,
    edge_weights,
):
    return array.mean(xp, river_network, field, locations, node_weights, edge_weights)


@multi_backend(allow_jax_jit=False)
def sum(
    xp,
    river_network,
    field,
    locations,
    node_weights,
    edge_weights,
):
    return array.sum(xp, river_network, field, locations, node_weights, edge_weights)


@multi_backend(allow_jax_jit=False)
def min(
    xp,
    river_network,
    field,
    locations,
    node_weights,
    edge_weights,
):
    return array.min(xp, river_network, field, locations, node_weights, edge_weights)


@multi_backend(allow_jax_jit=False)
def max(
    xp,
    river_network,
    field,
    locations,
    node_weights,
    edge_weights,
):
    return array.max(xp, river_network, field, locations, node_weights, edge_weights)



import json
from io import BytesIO
from urllib.request import Request, urlopen

import joblib

from earthkit.hydro._readers import (  # cache, from_grit,
    find_main_var,
    from_cama_nextxy,
    from_d8,
    import_earthkit_or_prompt_install,
)
from earthkit.hydro._utils.coords import get_core_grid_dims
from earthkit.hydro._utils.readers import from_file
from earthkit.hydro._version import __version__ as ekh_version
from earthkit.hydro.data_structures._network import RiverNetwork

from ._cache import cache

# read in major version
# if dev version, try add +1 to major version
# i.e. 0.1.dev90+gfdf4e33.d20250107 -> 1
# i.e. 0.1.0 -> 0
ekh_version = (
    int(ekh_version.split(".")[0]) + 1
    if "dev" in ekh_version
    else int(ekh_version.split(".")[0])
)


@cache
def create(
    path,
    river_network_format,
    source="file",
    use_cache=True,
    cache_dir=None,
    cache_fname="{ekh_version}_{hash}.joblib",
    cache_compression=1,
):
    """
    Creates a river network from the given path, format, and source.

    Parameters
    ----------
    path : str
        The path to the river network data. All common file formats are supported such
        as netCDF, GRIB, GeoTIFF, zarr, etc.
    river_network_format : str
        The format of the river network data.
        Supported formats are "precomputed", "cama", "pcr_d8", "esri_d8"
        and "merit_d8".
    source : str
        The source of the river network data. Default is `'file'`.
        For possible sources see:
        https://earthkit-data.readthedocs.io/en/latest/guide/sources.html.
    use_cache : bool, optional
        Whether to cache the loaded/created river network for quicker reloading. Default is True.
    cache_dir : str, optional
        Where to store the cached river networks. Default is None, which uses `tempfile.mkdtemp(suffix="_earthkit_hydro")`.
    cache_fname : str, optional
        A string template for the cache filename convention.
    cache_compression : int, optional
        A compression factor for the cached files.

    Returns
    -------
    RiverNetwork
        The river network object created from the given data.
    """
    if river_network_format == "precomputed":
        if source == "file":
            river_network_storage = joblib.load(path)
        elif source == "url":
            river_network_storage = joblib.load(BytesIO(urlopen(path).read()))
        else:
            raise ValueError(
                "Unsupported source for river network format"
                f"{river_network_format}: {source}."
            )
    elif river_network_format == "cama":
        ekd = import_earthkit_or_prompt_install(river_network_format, source)
        data = ekd.from_source(source, path).to_xarray(mask_and_scale=False)
        x, y = data.nextx.values, data.nexty.values
        river_network_storage = from_cama_nextxy(x, y)
        coord1, coord2 = get_core_grid_dims(data)
        river_network_storage.coords = {
            coord1: data[coord1].values,
            coord2: data[coord2].values,
        }
    elif (
        river_network_format == "pcr_d8"
        or river_network_format == "esri_d8"
        or river_network_format == "merit_d8"
    ):
        if path.endswith(".map"):
            data = from_file(path, mask=False)
            river_network_storage = from_d8(
                data, river_network_format=river_network_format
            )
            # coords not available
        else:
            ekd = import_earthkit_or_prompt_install(river_network_format, source)
            data = ekd.from_source(source, path).to_xarray(mask_and_scale=False)
            coord1, coord2 = get_core_grid_dims(data)
            var_name = find_main_var(data)
            river_network_storage = from_d8(
                data[var_name].values, river_network_format=river_network_format
            )
            river_network_storage.coords = {
                coord1: data[coord1].values,
                coord2: data[coord2].values,
            }
    # elif river_network_format == "grit":
    #     assert path.endswith(".gpkg")
    #     river_network_storage = from_grit(path)
    else:
        raise ValueError(f"Unsupported river network format: {river_network_format}.")

    return RiverNetwork(river_network_storage)


def load(
    domain,
    river_network_version,
    data_source=(
        "https://github.com/ecmwf/earthkit-hydro-store/raw/refs/heads/main/"
        "{ekh_version}/{domain}/{river_network_version}/river_network.joblib"
    ),
    *args,
    **kwargs,
):
    """
    Load a precomputed river network from a named domain and
    river_network_version.

    Supported networks are as follows:

    +----------------------+-----------+---------------------+----------------------------+----------------+
    | `domain`             | `version` | Details             | Note                       | Attribution    |
    +======================+===========+=====================+============================+================+
    | "efas"               | "5"       | 1arcmin European    |                            | [1]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "efas"               | "4"       | 5km European        | Smaller domain than v5     | [1]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "glofas"             | "4"       | 3arcmin global      | 60° South to 90° North     | [2]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "glofas"             | "3"       | 6arcmin global      | 60° South to 90° North     | [2]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "cama_03min"         | "4"       | 3arcmin global      |                            | [3]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "cama_05min"         | "4"       | 5arcmin global      |                            | [3]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "cama_06min"         | "4"       | 6arcmin global      |                            | [3]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "cama_15min"         | "4"       | 15arcmin global     |                            | [3]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "hydrosheds_05min"   | "1"       | 5arcmin global      | 56° South to 84° North     | [4]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+
    | "hydrosheds_06min"   | "1"       | 6arcmin global      | 56° South to 84° North     | [4]_           |
    +----------------------+-----------+---------------------+----------------------------+----------------+


    Parameters
    ----------
    domain : str
        The domain of the river network.
    river_network_version : str
        The version of the river network on the specified domain.
    data_source : str, optional
        The data source URL template for the river network.
    *args : tuple
        Additional positional arguments to pass to `create_river_network`.
    **kwargs : dict
        Additional keyword arguments to pass to `create_river_network`.

    Returns
    -------
    RiverNetwork
        The loaded river network.

    References
    ----------
    .. [1] Choulga, Margarita; Moschini, Francesca; Mazzetti, Cinzia; Grimaldi, Stefania; Disperati, Juliana; Beck, Hylke; Salamon, Peter; Prudhomme, Christel (2023): LISFLOOD static and parameter maps for Europe. European Commission, Joint Research Centre (JRC) [Dataset] PID: http://data.europa.eu/89h/f572c443-7466-4adf-87aa-c0847a169f23
    .. [2] Choulga, Margarita; Moschini, Francesca; Mazzetti, Cinzia; Disperati, Juliana; Grimaldi, Stefania; Beck, Hylke; Salamon, Peter; Prudhomme, Christel (2023): LISFLOOD static and parameter maps for GloFAS. European Commission, Joint Research Centre (JRC) [Dataset] PID: http://data.europa.eu/89h/68050d73-9c06-499c-a441-dc5053cb0c86
    .. [3] Yamazaki, Dai; Ikeshima, Daiki; Sosa, Jeison; Bates, Paul D.; Allen, George H.; Pavelsky, Tamlin M. (2019): MERIT Hydro: A high-resolution global hydrography map based on latest topography datasets. Water Resources Research, vol.55, pp.5053-5073, 2019, DOI: 10.1029/2019WR024873
    .. [4] Lehner, Bernhard; Verdin, Kristine; Jarvis, Andy (2008): New global hydrography derived from spaceborne elevation data. Eos, Transactions, 89(10): 93-94. Data available at https://www.hydrosheds.org.
    """

    try:
        uri = data_source.format(
            ekh_version=ekh_version,
            domain=domain,
            river_network_version=river_network_version,
        )
        network = create(uri, "precomputed", "url", *args, **kwargs)
    except Exception:
        uri = data_source.format(
            ekh_version=ekh_version - 1,
            domain=domain,
            river_network_version=river_network_version,
        )
        network = create(uri, "precomputed", "url", *args, **kwargs)

    return network


def available(
    data_source="https://api.github.com/repos/ecmwf/earthkit-hydro-store/git",
    token=None,
):
    """
    Prints the available precomputed networks.

    Parameters
    ----------
    data_source : str, optional
        Base github URI to query from.
    token : str, optional
        Github access token.
    """

    base_headers = {
        "Accept": "application/vnd.github+json",
        "User-Agent": "python-urllib-client",
    }
    if token:
        base_headers["Authorization"] = f"Bearer {token}"

    def github_api_request(url):
        req = Request(url, headers=base_headers)
        with urlopen(req) as response:
            if response.status != 200:
                raise Exception(f"GitHub API error {response.status}")
            return json.loads(response.read().decode())

    # get commit sha for main branch
    commit_sha = github_api_request(f"{data_source}/refs/heads/main")["object"]["sha"]

    # get entire tree
    tree_sha = github_api_request(f"{data_source}/commits/{commit_sha}")["tree"]["sha"]

    # get recursive tree
    tree_data = github_api_request(f"{data_source}/trees/{tree_sha}?recursive=1")[
        "tree"
    ]

    def is_valid_path(obj):
        return (
            ".joblib" in obj["path"].split("/")[-1]
            and ekh_version in obj["path"].split("/")[0]
        )

    print(
        "Available precomputed networks are:\n",
        *[
            '`ekh.river_network.load("{0}", "{1}")`\n'.format(
                *obj["path"].split("/")[1:3]
            )
            for obj in tree_data
            if is_valid_path(obj)
        ],
    )



from ._move import move_python as flow
from .metrics import metrics_func_finder


def calculate_move_metric(
    xp,
    river_network,
    field,
    metric,
    node_weights,
    edge_weights,
    flow_direction,
):
    if flow_direction == "up":
        invert_graph = True
        node_modifier_use_upstream = True
    elif flow_direction == "down":
        invert_graph = False
        node_modifier_use_upstream = True
    else:
        raise ValueError(
            f"flow_direction must be 'up' or 'down', got {flow_direction}."
        )

    if node_weights is None:
        if metric == "mean" or metric == "std" or metric == "var":
            node_weights = xp.ones(river_network.n_nodes, dtype=xp.float64)
    else:
        node_weights = xp.copy(node_weights)

    if edge_weights is not None:
        edge_weights = xp.copy(edge_weights)

    func = metrics_func_finder(metric, xp).func

    weighted_field = flow(
        xp,
        river_network,
        xp.zeros(field.shape),
        func,
        invert_graph,
        node_additive_weight=field if node_weights is None else field * node_weights,
        node_modifier_use_upstream=node_modifier_use_upstream,
        edge_multiplicative_weight=edge_weights,
    )

    if metric == "mean" or metric == "std" or metric == "var":
        counts = flow(
            xp,
            river_network,
            xp.zeros(field.shape),
            func,
            invert_graph,
            node_additive_weight=xp.copy(node_weights),
            node_modifier_use_upstream=node_modifier_use_upstream,
            edge_multiplicative_weight=edge_weights,
        )

        if metric == "mean":
            weighted_field /= counts
            return weighted_field
        elif metric == "var" or metric == "std":
            weighted_sum_of_squares = flow(
                xp,
                river_network,
                xp.zeros(field.shape),
                func,
                invert_graph,
                node_additive_weight=(
                    field**2 if node_weights is None else field**2 * node_weights
                ),
                node_modifier_use_upstream=node_modifier_use_upstream,
                edge_multiplicative_weight=edge_weights,
            )
            mean = weighted_field / counts
            weighted_sum_of_squares = weighted_sum_of_squares / counts - mean**2
            if metric == "var":
                return weighted_sum_of_squares
            elif metric == "std":
                return xp.sqrt(weighted_sum_of_squares)
    else:
        return weighted_field



import array_api_compat.numpy as np

from .array_backend import ArrayBackend


class NumPyBackend(ArrayBackend):
    def __init__(self):
        super().__init__(np)

    @property
    def name(self):
        return "numpy"

    def copy(self, x):
        return x.copy()

    def gather(self, arr, indices, axis=-1):
        assert axis == -1
        return arr[..., indices]

    def scatter_assign(self, target, indices, updates):
        target[..., indices] = updates
        return target

    def scatter_add(self, target, indices, updates):
        np.add.at(target, (*[slice(None)] * (target.ndim - 1), indices), updates)
        return target

    def scatter_max(self, target, indices, updates):
        np.maximum.at(target, (*[slice(None)] * (target.ndim - 1), indices), updates)
        return target

    def scatter_min(self, target, indices, updates):
        np.minimum.at(target, (*[slice(None)] * (target.ndim - 1), indices), updates)
        return target



from earthkit.hydro._core.accumulate import flow_downstream, flow_upstream
from earthkit.hydro._core.metrics import metrics_func_finder


def min(xp, river_network, field, locations, upstream, downstream):

    func_obj = metrics_func_finder("min", xp)

    out = xp.full(river_network.n_nodes, func_obj.base_val)

    out[locations] = field[locations]

    func = func_obj.func

    if downstream:
        out = flow_downstream(
            xp,
            river_network,
            out,
            func,
            node_additive_weight=field,
            node_modifier_use_upstream=False,
        )
    if upstream:
        out = flow_upstream(
            xp,
            river_network,
            out,
            func,
            node_additive_weight=field,
            node_modifier_use_upstream=True,
        )

    return out


def max(xp, river_network, field, locations, upstream, downstream):

    func_obj = metrics_func_finder("min", xp)

    out = xp.full(river_network.n_nodes, func_obj.base_val)

    out[locations] = field[locations]

    func = func_obj.func

    if downstream:
        out = flow_downstream(
            xp,
            river_network,
            out,
            func,
            node_additive_weight=field,
            node_modifier_use_upstream=False,
        )
    if upstream:
        out = flow_upstream(
            xp,
            river_network,
            out,
            func,
            node_additive_weight=field,
            node_modifier_use_upstream=True,
        )

    return out



class ArrayBackend:
    def __init__(self, module):
        self._mod = module

    def __getattr__(self, name):
        return getattr(self._mod, name)  # Delegate to underlying module

    @property
    def name(self):
        raise NotImplementedError

    def copy(self, x):
        raise NotImplementedError

    # extended functionality
    def gather(self, arr, indices, axis=-1):
        raise NotImplementedError

    def scatter_assign(self, target, indices, updates):
        raise NotImplementedError

    def scatter_add(self, target, indices, updates):
        raise NotImplementedError

    def scatter_max(self, target, indices, updates):
        raise NotImplementedError

    def scatter_min(self, target, indices, updates):
        raise NotImplementedError

    def scatter_mul(self, target, indices, updates):
        raise NotImplementedError



import numpy as np


def locations_to_1d(xp, river_network, locations):

    orig_locations = locations
    dict_locations = isinstance(locations, dict)
    if dict_locations:

        coord1_network_vals, coord2_network_vals = river_network.coords.values()

        locations = []
        if river_network.shape is None:  # vector network
            for coord1_val, coord2_val in orig_locations.values():
                indx = (
                    (coord1_val - coord1_network_vals) ** 2
                    + (coord2_val - coord2_network_vals) ** 2
                ).argmin()
                locations.append(int(indx))
        else:
            for coord1_val, coord2_val in orig_locations.values():
                indx = np.argmin((coord1_val - coord1_network_vals) ** 2)
                indy = np.argmin((coord2_val - coord2_network_vals) ** 2)
                locations.append((int(indx), int(indy)))

    locations = xp.asarray(locations, device=river_network.device)
    stations = locations

    if stations.ndim == 2 and stations.shape[1] == 2:
        if xp.name not in ["numpy", "cupy"]:
            raise NotImplementedError
        # TODO: make this code actually xp agnostic
        rows, cols = stations[:, 0], stations[:, 1]
        flat_indices = rows * river_network.shape[1] + cols
        flat_mask = river_network.mask
        reverse_map = -xp.ones(
            river_network.shape[0] * river_network.shape[1], dtype=int
        )
        reverse_map[flat_mask] = xp.arange(flat_mask.shape[0])
        masked_indices = reverse_map[flat_indices]
        if xp.any(masked_indices < 0):
            raise ValueError(
                "Some station points are not included in the masked array."
            )
        stations = xp.asarray(masked_indices, device=river_network.device)
    else:
        assert stations.ndim == 1

    return stations, locations, orig_locations



import os

import numpy as np


def compute_topological_labels(sources, sinks, downstream_nodes, n_nodes):

    use_rust = int(os.environ.get("USE_RUST", "-1"))

    if use_rust == 0:
        func = compute_topological_labels_python
    elif use_rust == 1:
        from earthkit.hydro._rust import compute_topological_labels_rust as func
    else:
        try:
            from earthkit.hydro._rust import compute_topological_labels_rust as func
        except (ModuleNotFoundError, ImportError):
            func = compute_topological_labels_python

    return func(sources, sinks, downstream_nodes, n_nodes)


def compute_topological_labels_python(
    sources: np.ndarray, sinks: np.ndarray, downstream_nodes: np.ndarray, n_nodes: int
):
    n_nodes = downstream_nodes.shape[0]
    inlets = downstream_nodes[sources]
    labels = np.zeros(n_nodes, dtype=np.intp)

    for n in range(1, n_nodes + 1):
        inlets = np.unique(inlets[inlets != n_nodes])  # subset to valid nodes
        if inlets.shape[0] == 0:
            break
        labels[inlets] = n  # update furthest distance from source
        inlets = downstream_nodes[inlets]

    if inlets.shape[0] != 0:
        raise ValueError("River Network contains a cycle.")
    labels[sinks] = n - 1  # put all sinks in last group in topological ordering

    return labels



import earthkit.hydro.move.array.__operations as array
from earthkit.hydro._utils.decorators import mask, multi_backend


@multi_backend()
def upstream(xp, river_network, field, node_weights, edge_weights, metric, return_type):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_func = mask(return_type == "gridded")(array.upstream)
    return decorated_func(xp, river_network, field, node_weights, edge_weights, metric)


@multi_backend()
def downstream(
    xp, river_network, field, node_weights, edge_weights, metric, return_type
):
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_func = mask(return_type == "gridded")(array.downstream)
    return decorated_func(xp, river_network, field, node_weights, edge_weights, metric)



from earthkit.hydro._utils.decorators import mask, multi_backend
from earthkit.hydro._utils.locations import locations_to_1d
from earthkit.hydro.distance.array import __operations as _operations


@multi_backend()
def min(xp, river_network, field, locations, upstream, downstream, return_type):
    if field is None:
        field = xp.ones(river_network.n_edges)
    locations, _, _ = locations_to_1d(xp, river_network, locations)
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_func = mask(return_type == "gridded")(_operations.min)
    return decorated_func(xp, river_network, field, locations, upstream, downstream)


@multi_backend()
def max(xp, river_network, field, locations, upstream, downstream, return_type):
    if field is None:
        field = xp.ones(river_network.n_edges)
    locations, _, _ = locations_to_1d(xp, river_network, locations)
    return_type = river_network.return_type if return_type is None else return_type
    if return_type not in ["gridded", "masked"]:
        raise ValueError("return_type must be either 'gridded' or 'masked'.")
    decorated_func = mask(return_type == "gridded")(_operations.max)
    return decorated_func(xp, river_network, field, locations, upstream, downstream)


def to_source(*args, **kwargs):
    raise NotImplementedError


def to_sink(*args, **kwargs):
    raise NotImplementedError



from functools import wraps

from earthkit.hydro._backends.find import get_array_backend


def multi_backend(allow_jax_jit=True):
    def decorator(func):
        compiled_jax_fn = None

        @wraps(func)
        def wrapper(river_network, *args, **kwargs):
            xp = get_array_backend(river_network.groups[0])
            backend_name = xp.name
            if backend_name == "jax" and allow_jax_jit:
                nonlocal compiled_jax_fn
                if compiled_jax_fn is None:
                    from jax import jit

                    def jax_func(*args, **kwargs):
                        return func(xp, river_network, *args, **kwargs)

                    compiled_jax_fn = jit(jax_func)
                return compiled_jax_fn(*args, **kwargs)
            else:
                return func(xp, river_network, *args, **kwargs)

        return wrapper

    return decorator



from earthkit.hydro._core.flow import propagate


def _flow_find(
    xp,
    river_network,
    field,
    overwrite=True,
    invert_graph=True,
):
    op = _find_catchments

    def operation(
        field,
        did,
        uid,
        eid,
    ):
        return op(
            xp,
            field,
            did,
            uid,
            eid,
            overwrite=overwrite,
        )

    field = propagate(
        river_network,
        river_network.groups,
        field,
        invert_graph,
        operation,
    )

    return field


def _find_catchments(xp, field, did, uid, eid, overwrite):
    """
    Updates field in-place with the value of its downstream nodes,
    dealing with missing values for 2D fields.

    Parameters
    ----------
    river_network : earthkit.hydro.network.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    grouping : numpy.ndarray
        The array of node indices.
    overwrite : bool
        If True, overwrite existing non-missing values in the field array.

    Returns
    -------
    None
    """
    down_not_missing = ~xp.isnan(xp.gather(field, uid, axis=-1))
    did = did[
        down_not_missing
    ]  # only update nodes where the downstream belongs to a catchment
    if not overwrite:
        up_is_missing = xp.isnan(xp.gather(field, did, axis=-1))
        did = did[up_is_missing]
    else:
        up_is_missing = None
    uid = (
        uid[down_not_missing][up_is_missing]
        if up_is_missing is not None
        else uid[down_not_missing]
    )
    updates = xp.gather(field, uid, axis=-1)
    return xp.scatter_assign(field, did, updates)



def metrics_func_finder(metric, xp):

    class SumBased:
        func = xp.scatter_add
        base_val = 0

    class MaxBased:
        func = xp.scatter_max
        base_val = -xp.inf

    class MinBased:
        func = xp.scatter_min
        base_val = xp.inf

    metrics_dict = {
        "sum": SumBased,
        "mean": SumBased,
        "std": SumBased,
        "var": SumBased,
        "max": MaxBased,
        "min": MinBased,
    }
    return metrics_dict[metric]
